import json
from pathlib import Path

import pytest
from unittest.mock import MagicMock, patch, call, PropertyMock, Mock
import asyncio
import railtracks as rt
from railtracks import Session, session

# ================= START Mock Fixture ============
@pytest.fixture
def mock_dependencies(monkeypatch):
    m_mark_session_logging_override = MagicMock()
    m_restore_module_logging = MagicMock()
    m_get_global_config = MagicMock()
    m_RTPublisher = MagicMock()
    m_ExecutionInfo = MagicMock(create_new=MagicMock())
    m_Coordinator = MagicMock()
    m_RTState = MagicMock()
    m_register_globals = MagicMock()
    m_delete_globals = MagicMock()

    monkeypatch.setattr('railtracks._session.mark_session_logging_override', m_mark_session_logging_override)
    monkeypatch.setattr('railtracks._session.restore_module_logging', m_restore_module_logging)
    monkeypatch.setattr('railtracks._session.get_global_config', m_get_global_config)
    monkeypatch.setattr('railtracks._session.RTPublisher', m_RTPublisher)
    monkeypatch.setattr('railtracks._session.ExecutionInfo', m_ExecutionInfo)
    monkeypatch.setattr('railtracks._session.Coordinator', m_Coordinator)
    monkeypatch.setattr('railtracks._session.RTState', m_RTState)
    monkeypatch.setattr('railtracks._session.register_globals', m_register_globals)
    monkeypatch.setattr('railtracks._session.delete_globals', m_delete_globals)

    return {
        'mark_session_logging_override': m_mark_session_logging_override,
        'restore_module_logging': m_restore_module_logging,
        'get_global_config': m_get_global_config,
        'RTPublisher': m_RTPublisher,
        'ExecutionInfo': m_ExecutionInfo,
        'Coordinator': m_Coordinator,
        'RTState': m_RTState,
        'register_globals': m_register_globals,
        'delete_globals': m_delete_globals,
    }
# ================ END Mock Fixture ===============

# ================= START Session: Construction & Context Manager ============
def test_runner_construction_with_explicit_config_and_context(mock_dependencies):
    context = {'foo': 'bar'}
    # Setup mocks with needed API
    pub_mock = mock_dependencies['RTPublisher'].return_value
    state_mock = mock_dependencies['RTState'].return_value
    info_mock = MagicMock()
    state_mock.info = info_mock

    # Should not raise
    r = Session(context=context)
    assert hasattr(r, 'publisher')
    assert hasattr(r, 'rt_state')
    assert hasattr(r, 'coordinator')
    assert r.rt_state.info == info_mock

def test_runner_construction_with_defaults(mock_dependencies):
    # Should call get_global_config()
    Session()
    assert mock_dependencies['get_global_config'].called

def test_runner_context_manager_closes_on_exit(mock_dependencies):

    context = {}
    runner = Session(context=context)
    with patch.object(runner, "_close") as mock_close:
        with runner:
            pass
        mock_close.assert_called_once()

# ================ END Session: Construction & Context Manager ===============


# ================= START Session: Singleton/Instance Id Behavior ============

def test_session_name_is_taken_from_executor_config():
    name = "abc123"

    r = Session(name=name)
    assert r.name == name

# ================ END Session: Singleton/Instance Id Behavior ===============


# ================= START Session: Logging Override Tests ===============

def test_session_with_logging_setting_marks_override(mock_dependencies):
    """Test that Session with logging_setting marks session logging override."""
    runner = Session(logging_setting="VERBOSE")
    assert mock_dependencies['mark_session_logging_override'].called
    assert runner._has_custom_logging is True

def test_session_with_log_file_marks_override(mock_dependencies):
    """Test that Session with log_file marks session logging override."""
    runner = Session(log_file="/tmp/test.log")
    assert mock_dependencies['mark_session_logging_override'].called
    assert runner._has_custom_logging is True

def test_session_with_both_logging_params_marks_override(mock_dependencies):
    """Test that Session with both logging params marks session logging override."""
    runner = Session(logging_setting="QUIET", log_file="/tmp/test.log")
    assert mock_dependencies['mark_session_logging_override'].called
    assert runner._has_custom_logging is True

def test_session_without_logging_params_no_override(mock_dependencies):
    """Test that Session without logging params does not mark override."""
    runner = Session()
    assert not mock_dependencies['mark_session_logging_override'].called
    assert runner._has_custom_logging is False

# ================ END Session: Logging Override Tests ===============


# ================= START Session: setup_subscriber ===============

def test_setup_subscriber_adds_subscriber_if_present():
    sub_subscriber = Mock()
    runner = Session(broadcast_callback=sub_subscriber)
    runner.publisher = MagicMock()
    with patch('railtracks._session.stream_subscriber', return_value="fake_stream_sub") as m_stream:
        runner._setup_subscriber()
        runner.publisher.subscribe.assert_called_once_with(
            "fake_stream_sub", name="Streaming Subscriber"
        )
        m_stream.assert_called_once_with(sub_subscriber)

def test_setup_subscriber_noop_if_no_subscriber(mock_dependencies):
    runner = Session()
    runner.executor_config.subscriber = None
    runner.publisher = MagicMock()
    with patch('railtracks._session.stream_subscriber') as m_stream:
        runner._setup_subscriber()
        runner.publisher.subscribe.assert_not_called()
        m_stream.assert_not_called()

# ================ END Session: setup_subscriber ===============


# ================= START Session: _close & __exit__ ===============

def test_close_calls_shutdown_and_delete(mock_dependencies):
    """Test that _close calls shutdown and delete_globals."""
    runner = Session()
    runner.rt_state = MagicMock()
    runner._close()
    assert runner.rt_state.shutdown.called
    assert mock_dependencies['delete_globals'].called

def test_close_restores_logging_when_custom_logging_used(mock_dependencies):
    """Test that _close calls restore_module_logging when custom logging was used."""
    runner = Session(logging_setting="VERBOSE")
    runner.rt_state = MagicMock()
    runner._close()
    assert mock_dependencies['restore_module_logging'].called

def test_close_does_not_restore_logging_when_no_custom_logging(mock_dependencies):
    """Test that _close does not call restore_module_logging when no custom logging."""
    runner = Session()
    runner.rt_state = MagicMock()
    runner._close()
    assert not mock_dependencies['restore_module_logging'].called

# ================ END Session: _close & __exit__ ===============


# ================= START Session: info property ===============

def test_info_property_returns_rt_state_info(mock_dependencies):
    runner = Session()
    rt_info = MagicMock()
    runner.rt_state.info = rt_info
    assert runner.info is rt_info

# ================ END Session: info property ===============


# ================= START Session: Check saved data ===============
def test_session_saves_data():
    name = "abs53562j12h267"

    serialization_mock = {"Key": "Value"}
    info = MagicMock()
    info.graph_serialization.return_value = serialization_mock


    with patch.object(Session, 'info', new_callable=PropertyMock) as mock_runner:
        mock_runner.return_value.graph_serialization.return_value = serialization_mock

        r = Session(
            name=name,
            save_state=True,
        )
        r.__exit__(None, None, None)


    path = Path(".railtracks/data/sessions") / f"{r.name}_{r._identifier}.json"
    data = json.loads(path.read_text())
    assert data["runs"] == serialization_mock
    assert "session_id" in data
    assert data["session_name"] == name


def test_session_not_saves_data():
    config = MagicMock()

    run_id = "Run 2"
    config.run_identifier = run_id
    config.save_state = False

    serialization_mock = '{"Key": "Value"}'
    info = MagicMock()
    info.graph_serialization.return_value = serialization_mock

    with patch.object(Session, 'info', new_callable=PropertyMock) as mock_runner:
        mock_runner.return_value.graph_serialization.return_value = serialization_mock

        r = Session(name=run_id, save_state=False)
        r.__exit__(None, None, None)




    path = Path(".railtracks/data/sessions") / f"{run_id}.json"
    assert not path.is_file()


def test_session_fallback_on_invalid_name():
    """Test that session falls back to identifier-only filename when name causes issues."""
    # Use a name that would cause issues in file path creation
    invalid_name = "test/invalid:name*with|bad<chars>"

    serialization_mock = {"Key": "Value"}

    with patch.object(Session, 'info', new_callable=PropertyMock) as mock_runner:
        mock_runner.return_value.graph_serialization.return_value = serialization_mock

        # Mock Path.touch to raise an exception when the path contains the invalid name in the filename
        original_touch = Path.touch
        def mock_touch(self, *args, **kwargs):
            # Only raise exception if the invalid name is in the filename part (not just any path)
            if invalid_name in self.name:
                raise OSError("Invalid characters in filename")
            return original_touch(self, *args, **kwargs)

        with patch.object(Path, 'touch', mock_touch), \
             patch('railtracks._session.logger') as mock_logger:

            r = Session(name=invalid_name, save_state=True)
            r.__exit__(None, None, None)

            # Verify that a warning was logged about the invalid name
            mock_logger.warning.assert_called()
            warning_calls = [call[0][0] for call in mock_logger.warning.call_args_list]
            fallback_warning = None
            for call in warning_calls:
                if "falling back to using the unique identifier only" in call:
                    fallback_warning = call
                    break

            assert fallback_warning is not None, "Expected fallback warning not found"
            assert invalid_name in fallback_warning

            # Verify that the fallback file was created (identifier only)
            fallback_path = Path(".railtracks/data/sessions") / f"{r._identifier}.json"
            assert fallback_path.exists()

            # Clean up
            if fallback_path.exists():
                fallback_path.unlink()


# ================= START Session: Decorator Tests ===============

def test_session_decorator_creates_function():
    """Test that the session decorator returns a decorator function."""
    decorator = session()
    assert callable(decorator)

def test_session_decorator_with_parameters():
    """Test session decorator with various parameters."""
    decorator = session(
        timeout=30,
        context={"test": "value"},
        end_on_error=True,
        logging_setting="VERBOSE"
    )
    assert callable(decorator)

@pytest.mark.asyncio
async def test_session_decorator_wraps_async_function(mock_dependencies):
    """Test that the decorator properly wraps an async function and returns tuple."""
    @session(timeout=10)
    async def test_function():
        return "test_result"

    result, session_obj = await test_function()
    assert result == "test_result"
    assert isinstance(session_obj, Session)

@pytest.mark.asyncio
async def test_session_decorator_with_function_args(mock_dependencies):
    """Test that the decorator preserves function arguments and returns tuple."""
    @session()
    async def test_function(arg1, arg2, kwarg1=None):
        return f"{arg1}-{arg2}-{kwarg1}"

    result, session_obj = await test_function("a", "b", kwarg1="c")
    assert result == "a-b-c"
    assert isinstance(session_obj, Session)

@pytest.mark.asyncio
async def test_session_decorator_context_manager_behavior(mock_dependencies):
    """Test that the decorator properly manages session lifecycle."""
    session_created = False
    session_closed = False

    original_init = Session.__init__
    original_exit = Session.__exit__

    def mock_init(self, *args, **kwargs):
        nonlocal session_created
        session_created = True
        return original_init(self, *args, **kwargs)

    def mock_exit(self, *args, **kwargs):
        nonlocal session_closed
        session_closed = True
        return original_exit(self, *args, **kwargs)

    with patch.object(Session, '__init__', mock_init), \
         patch.object(Session, '__exit__', mock_exit):

        @session()
        async def test_function():
            return "done"

        result, session_obj = await test_function()
        assert result == "done"
        assert isinstance(session_obj, Session)

    assert session_created
    assert session_closed

def test_session_decorator_raises_error_on_sync_function():
    """Test that the session decorator raises TypeError when applied to sync function."""
    with pytest.raises(TypeError, match="@session decorator can only be applied to async functions"):
        @session()
        def sync_function():
            return "this should fail"

def test_session_decorator_error_message_contains_function_name():
    """Test that the error message includes the function name."""
    with pytest.raises(TypeError, match="Function 'my_sync_func' is not async"):
        @session()
        def my_sync_func():
            return "this should fail"

def test_rt_session_decorator_raises_error_on_sync_function():
    """Test that rt.session also raises TypeError when applied to sync function."""
    with pytest.raises(TypeError, match="@session decorator can only be applied to async functions"):
        @session()
        def sync_function():
            return "this should fail"

@pytest.mark.asyncio
async def test_session_decorator_returns_session_object(mock_dependencies):
    """Test that decorator returns both result and session object with access to session info."""
    @session(name="test-session-123")
    async def test_function():
        return "test_result"

    result, session_obj = await test_function()

    # Verify we get both the result and session
    assert result == "test_result"
    assert isinstance(session_obj, Session)
    assert session_obj.name == "test-session-123"

    # Verify we can access session properties
    assert hasattr(session_obj, 'info')
    assert hasattr(session_obj, 'payload')

@pytest.mark.asyncio
async def test_session_decorator_handles_tuple_returns(mock_dependencies):
    """Test that decorator properly handles functions that return tuples."""
    @session()
    async def function_returning_tuple():
        return "hello", 42, True

    result, session_obj = await function_returning_tuple()

    # The result should be the tuple returned by the function
    assert result == ("hello", 42, True)
    assert isinstance(session_obj, Session)

    # The tuple structure is preserved
    val1, val2, val3 = result
    assert val1 == "hello"
    assert val2 == 42
    assert val3 == True

# ================ END Session: Decorator Tests ===============



import pytest
from unittest.mock import Mock, MagicMock, patch
from railtracks.vector_stores.vector_store_base import Chunk, SearchResult, FetchResult
from railtracks.vector_stores.chroma import ChromaVectorStore


# ============================================================================
# Generic Test Data Fixtures (reusable across all vector store tests)
# ============================================================================

@pytest.fixture
def simple_vector():
    """Simple 3-dimensional vector for testing."""
    return [0.1, 0.2, 0.3]


@pytest.fixture
def short_vector():
    """Short vector for basic operations."""
    return [0.1, 0.2, 0.3, 0.4, 0.5]


@pytest.fixture
def long_vector():
    """Long vector (1000 dimensions) for testing scalability."""
    return [float(i) for i in range(1000)]


@pytest.fixture
def zero_vector():
    """Zero vector for distance tests."""
    return [0.0, 0.0, 0.0]


@pytest.fixture
def simple_metadata():
    """Simple metadata dictionary."""
    return {"key": "value"}


@pytest.fixture
def complex_metadata():
    """Complex metadata with multiple keys."""
    return {"author": "John", "date": "2024-01-01", "tags": "test,demo"}


@pytest.fixture
def custom_id():
    """Custom ID string for testing."""
    return "custom_id_123"


@pytest.fixture
def simple_content():
    """Simple content string."""
    return "Test content"


@pytest.fixture
def simple_document():
    """Simple document identifier."""
    return "test_doc.txt"


@pytest.fixture
def zero_distance():
    """Zero distance value."""
    return 0.0


@pytest.fixture
def small_distance():
    """Small distance value for similarity."""
    return 0.1


@pytest.fixture
def large_distance():
    """Large distance value for dissimilarity."""
    return 0.9


# ============================================================================
# Embedding Function Fixtures
# ============================================================================

@pytest.fixture
def mock_embedding_function():
    """Mock embedding function that returns consistent embeddings."""
    def embed(texts: list[str]) -> list[list[float]]:
        # Return embeddings with same length as input, consistent for testing
        return [[float(i) for i in range(5)] for _ in texts]
    return embed


@pytest.fixture
def mock_chroma_ephemeral_client():
    """Mock Chroma EphemeralClient."""
    client = MagicMock()
    collection = MagicMock()
    client.get_or_create_collection.return_value = collection
    return client, collection


@pytest.fixture
def mock_chroma_persistent_client():
    """Mock Chroma PersistentClient."""
    client = MagicMock()
    collection = MagicMock()
    client.get_or_create_collection.return_value = collection
    return client, collection


@pytest.fixture
def mock_chroma_http_client():
    """Mock Chroma HttpClient."""
    client = MagicMock()
    collection = MagicMock()
    client.get_or_create_collection.return_value = collection
    return client, collection


@pytest.fixture
def sample_chunk():
    """Create a sample Chunk for testing."""
    return Chunk(
        content="Test content",
        document="test_doc.txt",
        metadata={"category": "test"}
    )


@pytest.fixture
def sample_chunks():
    """Create multiple sample Chunks for testing."""
    return [
        Chunk(
            content="First test content",
            document="doc1.txt",
            metadata={"id": 1}
        ),
        Chunk(
            content="Second test content",
            document="doc2.txt",
            metadata={"id": 2}
        ),
        Chunk(
            content="Third test content",
            document="doc3.txt",
            metadata={"id": 3}
        ),
    ]


@pytest.fixture
def sample_search_results():
    """Create sample SearchResult objects."""
    return [
        SearchResult(
            id="id1",
            distance=0.1,
            content="Result 1",
            vector=[0.1, 0.2, 0.3, 0.4, 0.5],
            document="doc1.txt",
            metadata={"source": "doc1"}
        ),
        SearchResult(
            id="id2",
            distance=0.2,
            content="Result 2",
            vector=[0.2, 0.3, 0.4, 0.5, 0.6],
            document="doc2.txt",
            metadata={"source": "doc2"}
        ),
    ]


@pytest.fixture
def sample_fetch_results():
    """Create sample FetchResult objects."""
    return [
        FetchResult(
            id="id1",
            content="Fetched content 1",
            vector=[0.1, 0.2, 0.3, 0.4, 0.5],
            document="doc1.txt",
            metadata={"key": "value1"}
        ),
        FetchResult(
            id="id2",
            content="Fetched content 2",
            vector=[0.2, 0.3, 0.4, 0.5, 0.6],
            document="doc2.txt",
            metadata={"key": "value2"}
        ),
    ]


@pytest.fixture
def mock_chromadb_module():
    """Mock the chromadb module for import patching."""
    mock_chromadb = MagicMock()
    mock_chromadb.EphemeralClient = MagicMock()
    mock_chromadb.PersistentClient = MagicMock()
    mock_chromadb.HttpClient = MagicMock()
    return mock_chromadb


# ============================================================================
# Chroma Test Helpers
# ============================================================================
@pytest.fixture
def reset_chroma_class():
    def _reset_chroma_class():
        """Reset ChromaVectorStore._chroma to ensure clean initialization state.
        
        Use this in tests that call ChromaVectorStore.class_init() to ensure
        the class_init method actually runs instead of finding an existing client.
        """
        if hasattr(ChromaVectorStore, "_chroma"):
            delattr(ChromaVectorStore, "_chroma")
    
    return _reset_chroma_class


@pytest.fixture
def chroma_mocks():
    """Fixture that provides a clean ChromaVectorStore with mocked collection.
    
    Returns:
        dict: Contains 'client' (mock Chroma client) and 'collection' (mock collection)
    
    Example:
        def test_something(self, chroma_mocks):
            from railtracks.vector_stores.chroma import ChromaVectorStore
            ChromaVectorStore._chroma = chroma_mocks["client"]
            store = ChromaVectorStore(collection_name="test", ...)
    """
    mock_client = MagicMock()
    mock_collection = MagicMock()
    mock_client.get_or_create_collection.return_value = mock_collection
    return {"client": mock_client, "collection": mock_collection}


@pytest.fixture
def chroma_store(mock_embedding_function, chroma_mocks):
    """Fixture that provides a fully initialized ChromaVectorStore instance.
    
    This is the preferred way to test ChromaVectorStore methods. It gives you
    a real store instance with a mocked collection, so you only need to mock
    the specific methods you're testing.
    
    Args:
        mock_embedding_function: Fixture for embedding function
        chroma_mocks: Fixture containing mocked client and collection
    
    Returns:
        ChromaVectorStore: Fully initialized store with mocked collection
    
    Example:
        def test_upsert(self, chroma_store, chroma_mocks):
            # Store is already initialized! Just configure the mock for specific method.
            chroma_mocks["collection"].upsert.return_value = ["id1"]
            result = chroma_store.upsert("Test content")
            assert result == "id1"
    """
    from railtracks.vector_stores.chroma import ChromaVectorStore
    
    # Patch class_init to skip the actual Chroma client initialization
    with patch.object(ChromaVectorStore, "class_init"):
        # Set up the class-level _chroma to our mock
        ChromaVectorStore._chroma = chroma_mocks["client"]
        
        # Create a real store instance (but with patched class_init)
        store = ChromaVectorStore(
            collection_name="test_collection",
            embedding_function=mock_embedding_function,
            path="./test-data"
        )
        
        # Patch the instance's _collection to use our mock collection
        original_collection = store._collection
        store._collection = chroma_mocks["collection"]
        
        yield store
        
        # Restore original (cleanup)
        store._collection = original_collection



@pytest.fixture
def patch_all_chromadb_clients(*args):
    """Context manager that patches all chromadb client types simultaneously.
    
    This patches PersistentClient, HttpClient, and EphemeralClient in one call.
    The main client is the first patch (which can be asserted on), and the others
    are just patched to prevent accidental use.
    
    Yields:
        The first mock (typically the one being tested)
    
    Example:
        with patch_all_chromadb_clients("chromadb.PersistentClient") as mock_persistent:
            # Other clients are also patched
            reset_chroma_class()
            ChromaVectorStore.class_init(path="./test-data", host=None, port=None)
            mock_persistent.assert_called_once()
    """
    clients = ["chromadb.PersistentClient", "chromadb.HttpClient", "chromadb.EphemeralClient"]
    
    # Find which client is the main one (first arg)
    main_client = args[0] if args else "chromadb.PersistentClient"
    other_clients = [c for c in clients if c != main_client]
    
    with patch(main_client) as mock_main:
        with patch(other_clients[0]), patch(other_clients[1]):
            yield mock_main



import asyncio

import pytest
import railtracks as rt
from railtracks.llm import Message, MessageHistory
from railtracks.llm.response import Response
from railtracks.llm.message import Role


def test_prompt_injection(mock_llm):
    prompt = "{secret}"

    def return_message(messages: MessageHistory) -> Response:
        return Response(message=Message(role=Role.assistant, content=messages[-1].content))

    model = mock_llm()
    model._chat = return_message

    node = rt.agent_node(system_message=prompt, llm=model)

    async def top_level():
        with rt.Session(context={"secret": "tomato"}):
            response = await rt.call(node, user_input=MessageHistory())
        return response

    response = asyncio.run(top_level())
    assert response.content == "tomato"


def test_prompt_injection_bypass(mock_llm):
    prompt = "{{secret_value}}"

    def return_message(messages: MessageHistory) -> Response:
        return Response(message=Message(role=Role.assistant, content=messages[-1].content))

    model = mock_llm()
    model._chat = return_message

    node = rt.agent_node(system_message=prompt, llm=model)

    async def top_level():
        with rt.Session(context={"secret_value": "tomato"}):
            response = await rt.call(node, user_input=MessageHistory())
        return response

    response = asyncio.run(top_level())

    assert response.content == "{secret_value}"


def test_prompt_numerical(mock_llm):
    prompt = "{1}"

    def return_message(messages: MessageHistory) -> Response:
        return Response(message=Message(role=Role.assistant, content=messages[-1].content))

    model = mock_llm()
    model._chat = return_message

    node = rt.agent_node(
        system_message=prompt,
        llm=model
    )

    async def top_level():
        with rt.Session(context={"1": "tomato"}):
            response = await rt.call(node, user_input=MessageHistory())
        return response

    response = asyncio.run(top_level())

    assert response.content == "tomato"


def test_prompt_not_in_context(mock_llm):
    prompt = "{secret2}"

    def return_message(messages: MessageHistory) -> Response:
        return Response(message=Message(role=Role.assistant, content=messages[-1].content))

    model = mock_llm()
    model._chat = return_message

    node = rt.agent_node(
        system_message=prompt,
        llm=model
    )

    async def top_level():
        with rt.Session():
            response = await rt.call(node, user_input=MessageHistory())

        return response

    response = asyncio.run(top_level())

    assert response.content == "{secret2}"


@pytest.mark.order("last")
def test_prompt_injection_global_config_bypass(mock_llm):
    prompt = "{secret_value}"

    def return_message(messages: MessageHistory) -> Response:
        return Response(message=Message(role=Role.assistant, content=messages[-1].content))

    model = mock_llm()
    model._chat = return_message

    node = rt.agent_node(
        system_message=prompt,
        llm=model
    )

    async def top_level():
        with rt.Session(context={"secret_value": "tomato"}, prompt_injection=False):
            response = await rt.call(node, user_input=MessageHistory())

        return response

    response = asyncio.run(top_level())
    assert response.content == "{secret_value}"



from typing import Literal, TypeVar

from ._llm_base import StringOutputMixIn
from ._tool_call_base import OutputLessToolCallLLM, StreamingOutputLessToolCallLLM
from .response import StringResponse

_TStream = TypeVar("_TStream", Literal[True], Literal[False])


class ToolCallLLM(
    StringOutputMixIn,
    OutputLessToolCallLLM[StringResponse],
):
    pass


class StreamingToolCallLLM(
    StringOutputMixIn, StreamingOutputLessToolCallLLM[StringResponse]
):
    pass



import pytest
import litellm
from unittest.mock import patch

from railtracks.llm.models import AzureAILLM
from railtracks.llm._exception_base import RTLLMError
from railtracks.llm.models._litellm_wrapper import LiteLLMWrapper

from railtracks.llm.response import Response
from railtracks.llm.message import AssistantMessage


MODEL_NAME = "deepseek-r1"
TEST_CHAT_MODEL_NAME = "azure_ai/deepseek-r1"
TEST_TOOLCALLING_MODEL_NAME = "azure_ai/deepseek-v3-0324"


@pytest.fixture
def response():
    """Fixture to provide a mock response for testing"""
    return Response(
        message=AssistantMessage("This is a response with no tool calls."),
    )


def test_model_type():
    """Test if the model type is correctly returned"""
    assert AzureAILLM.model_gateway() == "AzureAI"


def test_init_success():
    """Test successful initialization of AzureAILLM"""
    llm = AzureAILLM(model_name=TEST_CHAT_MODEL_NAME)
    assert llm._model_name == TEST_CHAT_MODEL_NAME


def test_init_model_not_available():
    """Test initialization with a model that is not available"""
    with pytest.raises(RTLLMError, match="Model 'non_existent_model' is not available"):
        AzureAILLM(model_name="non_existent_model")


def test_chat_success(message_history):
    """Test successful chat response from AzureAILLM"""
    llm = AzureAILLM(model_name=TEST_CHAT_MODEL_NAME)

    with patch.object(
        litellm,
        "completion",
        return_value=litellm.utils.ModelResponse(
            choices=[{"message": {"content": "This is a response from Azure AI."}}]
        ),
    ):
        response = llm.chat(message_history)
        assert response.message.content == "This is a response from Azure AI."


def test_chat_failure(message_history):
    """Test handling of chat failure due to an internal server error"""
    llm = AzureAILLM(model_name=TEST_CHAT_MODEL_NAME)

    with patch.object(
        litellm,
        "completion",
        side_effect=litellm.InternalServerError(
            "Internal server error", "azure_ai", MODEL_NAME
        ),
    ):
        with pytest.raises(
            RTLLMError, match="Azure AI LLM error while processing the request"
        ):
            llm.chat(message_history)


def test_chat_with_tools_success(message_history, response, tool):
    """
    Test successful chat with tools response from AzureAILLM
    Note there is no need to test for actual tool invocations
    """
    llm = AzureAILLM(model_name=TEST_TOOLCALLING_MODEL_NAME)

    with patch.object(LiteLLMWrapper, "chat_with_tools", return_value=response):
        response = llm.chat_with_tools(message_history, [tool])
        assert response.message.content == "This is a response with no tool calls."


def test_chat_with_tools_failure(message_history, tool):
    """"""
    llm = AzureAILLM(model_name=TEST_CHAT_MODEL_NAME)

    with patch.object(litellm, "supports_function_calling", return_value=False):
        with pytest.raises(RTLLMError):
            llm.chat_with_tools(message_history, [tool])



import time
import threading
import pickle
import pytest

from railtracks.utils.profiling import Stamp, StampManager

# ================= START Fixtures & Helpers ==================
@pytest.fixture
def stampmanager_fixture():
    return StampManager()

@pytest.fixture
def stamp_fixture():
    return Stamp(time=12345, step=2, identifier="uniq")

# ================= END Fixtures & Helpers ====================


# ================= START Stamp Basic/Operator tests ==================

def test_stamp_lt_orders_by_step():
    """__lt__ gives precedence to step before time"""
    s1 = Stamp(time=10, step=1, identifier="foo")
    s2 = Stamp(time=4, step=2, identifier="foo")
    assert s1 < s2

def test_stamp_lt_orders_by_time_on_same_step():
    s1 = Stamp(time=10, step=1, identifier="foo")
    s2 = Stamp(time=15, step=1, identifier="foo")
    assert s1 < s2

def test_stamp_hash_is_unique_for_unique_fields():
    s1 = Stamp(time=10, step=1, identifier="foo")
    s2 = Stamp(time=10, step=1, identifier="bar")
    assert hash(s1) != hash(s2)

def test_stamp_hash_matches_tuple(stamp_fixture):
    s = stamp_fixture
    assert hash(s) == hash((s.time, s.step, s.identifier))

# ================= END Stamp Basic/Operator tests ====================



# ============ START StampManager single/parallel/main path tests ================

def test_single_stamper():
    sm = StampManager()
    message = "Hello world"
    t = time.time()
    stamp = sm.create_stamp(message)

    assert stamp.step == 0
    assert stamp.identifier == message
    assert t <= stamp.time <= time.time()

    logs = sm.step_logs
    assert len(logs) == 1
    assert logs[0] == [message] or logs[0][0] == message  # handle dict/list
    assert sm.all_stamps[0].identifier == message

def test_multi_stamp():
    sm = StampManager()
    message = "Hello world"
    t1 = time.time()
    stamp1 = sm.create_stamp(message)

    assert stamp1.step == 0
    assert stamp1.identifier == message
    assert t1 <= stamp1.time <= time.time()

    t2 = time.time()
    stamp2 = sm.create_stamp(message)

    assert stamp2.step == 1
    assert stamp2.identifier == message
    assert t2 <= stamp2.time <= time.time()

def test_parallel_stamps():
    sm = StampManager()

    stamp_gen = sm.stamp_creator()
    t = time.time()
    uno = stamp_gen("1")
    dos = stamp_gen("2")
    tres = stamp_gen("3")

    assert uno.step == 0
    assert dos.step == 0
    assert tres.step == 0

    assert uno.identifier == "1"
    assert dos.identifier == "2"
    assert tres.identifier == "3"

    assert t <= uno.time <= dos.time
    assert uno.time <= dos.time <= tres.time
    assert dos.time <= tres.time <= time.time()

def test_two_parallel_stamps():
    sm = StampManager()
    stamp_gen_1 = sm.stamp_creator()
    stamp_gen_2 = sm.stamp_creator()

    uno_1 = stamp_gen_1("1")
    uno_2 = stamp_gen_2("1")
    dos_1 = stamp_gen_1("2")
    dos_2 = stamp_gen_2("2")
    tres_1 = stamp_gen_1("3")
    tres_2 = stamp_gen_2("3")

    assert uno_1.step == 0
    assert uno_2.step == 1
    assert dos_1.step == 0
    assert dos_2.step == 1
    assert tres_1.step == 0
    assert tres_2.step == 1

    assert uno_1.identifier == "1"
    assert uno_2.identifier == "1"
    assert dos_1.identifier == "2"
    assert dos_2.identifier == "2"
    assert tres_1.identifier == "3"
    assert tres_2.identifier == "3"

    assert uno_1.time <= uno_2.time
    assert uno_2.time <= dos_1.time
    assert dos_1.time <= dos_2.time
    assert dos_2.time <= tres_1.time
    assert tres_1.time <= tres_2.time

    assert uno_1 < uno_2
    assert dos_1 < dos_2

def test_combo():
    sm = StampManager()
    stamp_gen_1 = sm.stamp_creator()

    uno = stamp_gen_1("1")
    dos = stamp_gen_1("2")
    singleton = sm.create_stamp("3")
    tres = stamp_gen_1("4")

    assert uno.step == 0
    assert dos.step == 0
    assert singleton.step == 1
    assert tres.step == 0

    assert uno.identifier == "1"
    assert dos.identifier == "2"
    assert singleton.identifier == "3"
    assert tres.identifier == "4"

    assert uno.time <= dos.time
    assert dos.time <= singleton.time
    assert singleton.time <= tres.time

# ============= END StampManager single/parallel/main path tests ================

# ================ START StampManager save/state/lock/pickle ================

def test_save_state():
    sm = StampManager()
    stamp_gen_1 = sm.stamp_creator()
    uno = stamp_gen_1("1")

    state = sm.__getstate__()
    assert "_stamp_lock" not in state
    sm2 = StampManager()
    sm2.__setstate__(state)

    duo = sm2.create_stamp("2")
    assert duo.step == 1
    assert uno.step == 0

def test_create_lock_returns_new_lock():
    lock_type = type(threading.Lock())
    l1 = StampManager._create_lock()
    l2 = StampManager._create_lock()
    assert isinstance(l1, lock_type)
    assert l1 is not l2

def test_stampmanager_pickle_roundtrip(stampmanager_fixture):
    lock_type = type(threading.Lock())
    sm = stampmanager_fixture
    sm.create_stamp("abc")
    pkl = pickle.dumps(sm)
    sm2 = pickle.loads(pkl)
    # _stamp_lock should be re-created, not shared
    assert isinstance(sm2._stamp_lock, lock_type)
    assert sm2.step_logs == sm.step_logs

# ================ END StampManager save/state/lock/pickle ================

# ================= START StampManager convenience/getters =============
def test_create_stamp_basic(stampmanager_fixture):
    sm = stampmanager_fixture
    s = sm.create_stamp("first")
    assert isinstance(s, Stamp)
    assert s.step == 0
    assert s.identifier == "first"
    s2 = sm.create_stamp("second")
    assert s2.step == 1

def test_create_stamp_step_logs(stampmanager_fixture):
    sm = stampmanager_fixture
    sm.create_stamp("msg1")
    sm.create_stamp("msg2")
    logs = sm.step_logs
    assert logs[0] == ["msg1"]
    assert logs[1] == ["msg2"]

def test_stamp_creator_shared_step(stampmanager_fixture):
    sm = stampmanager_fixture
    creator = sm.stamp_creator()
    st1 = creator("a")
    st2 = creator("b")
    assert st1.step == st2.step
    assert st1.identifier == "a"
    assert st2.identifier == "b"
    assert sm._step >= st2.step + 1

def test_stamp_creator_step_logs(stampmanager_fixture):
    sm = stampmanager_fixture
    creator = sm.stamp_creator()
    creator("foo")
    creator("bar")
    logs = sm.step_logs
    log_vals = list(logs.values()) if isinstance(logs, dict) else logs
    assert "foo" in log_vals[0]
    assert "bar" in log_vals[0]

def test_all_stamps_sorted_returns_sorted(stampmanager_fixture):
    sm = stampmanager_fixture
    s1 = sm.create_stamp("msg1")
    time.sleep(0.01)
    s2 = sm.create_stamp("msg2")
    all_stamps = sm.all_stamps
    assert all_stamps[0].identifier == "msg1"
    assert all_stamps[1].identifier == "msg2"

def test_step_logs_is_deepcopy(stampmanager_fixture):
    sm = stampmanager_fixture
    sm.create_stamp("x")
    sl1 = sm.step_logs
    sl1[0].append("evil")
    assert sl1 != sm.step_logs

def test_all_stamps_is_deepcopy(stampmanager_fixture):
    sm = stampmanager_fixture
    sm.create_stamp("foo")
    all1 = sm.all_stamps
    all1[0].identifier = "evil"
    # mutated copy should not affect original
    assert all1 != sm.all_stamps

# ================ END StampManager convenience/getters ==============



from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING, Callable, Optional, TypeVar, Union, overload
from uuid import uuid4

from .chunking.base_chunker import Chunk
from .vector_store_base import (
    FetchResponse,
    FetchResult,
    MetadataKeys,
    SearchResponse,
    SearchResult,
    VectorStore,
)

if TYPE_CHECKING:
    from chromadb.api.types import Include
    from chromadb.base_types import Where, WhereDocument


CONTENT = MetadataKeys.CONTENT.value

T = TypeVar("T")

OneOrMany = Union[T, list[T]]


class ChromaVectorStore(VectorStore):
    """ChromaDB-backed implementation of :class:`VectorStore`.

    This class wraps a Chroma collection and translates between Chroma's
    API and the project's neutral types. The implementation currently
    supports upserting lists of either :class:`Chunk` or strings and
    querying by text strings.
    """

    @classmethod
    def class_init(
        cls, path: Optional[str], host: Optional[str], port: Optional[int]
    ) -> None:
        """Lazily initialize the shared Chroma client.

        This method performs an optional import of Chroma and creates a
        persistent, HTTP or ephemeral client depending on the parameters.

        Args:
            path: Filesystem path for a persistent client (optional).
            host: Hostname for an HTTP client (optional).
            port: Port for an HTTP client (optional).

        Raises:
            ImportError: If the `chromadb` package is not installed.
        """
        if not hasattr(cls, "_chroma"):
            try:
                import chromadb

                # Provide just a path for local store
                if path and not host and not port:
                    cls._chroma = chromadb.PersistentClient(path=path)
                # Provide just a host and port for http store
                elif not path and host and port:
                    cls._chroma = chromadb.HttpClient(host=host, port=port)
                # Provide nothing for temporary store
                elif not path and not host and not port:
                    cls._chroma = chromadb.EphemeralClient()
                else:
                    raise ValueError(
                        "Invalid combination of path, host, and port for Chroma client."
                    )
            except ImportError:
                raise ImportError(
                    "Chroma package is not installed. Please install railtracks[chroma]."
                )

    @overload
    def __init__(
        self,
        collection_name: str,
        embedding_function: Callable[[list[str]], list[list[float]]],
        *,
        path: str,
    ) -> None: ...

    @overload
    def __init__(
        self,
        collection_name: str,
        embedding_function: Callable[[list[str]], list[list[float]]],
        *,
        host: str,
        port: int,
    ) -> None: ...

    @overload
    def __init__(
        self,
        collection_name: str,
        embedding_function: Callable[[list[str]], list[list[float]]],
    ) -> None: ...

    def __init__(
        self,
        collection_name: str,
        embedding_function: Callable[[list[str]], list[list[float]]],
        path: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
    ):
        """Create a ChromaVectorStore instance.

        Args:
            collection_name: Name of the Chroma collection to use or create.
            embedding_function: Callable that maps a list of strings to a list
                of embedding vectors.
            path: Optional path for persistent Chroma storage.
            host: Optional HTTP host for remote Chroma.
            port: Optional HTTP port for remote Chroma.
        """
        self._collection_name = collection_name
        self._embedding_function = embedding_function

        ChromaVectorStore.class_init(path, host, port)
        self._collection = self._chroma.get_or_create_collection(collection_name)

    # In future should have our own chunking service so we can accept documents for users
    @overload
    def upsert(self, content: Chunk | str) -> str: ...

    @overload
    def upsert(self, content: list[Chunk] | list[str]) -> list[str]: ...

    def upsert(self, content: OneOrMany[Chunk] | OneOrMany[str]) -> OneOrMany[str]:
        """Upsert a batch of chunks or raw strings into the collection.

        The method accepts a list of :class:`Chunk` instances or plain strings.
        Each element is embedded via ``embedding_function`` and stored along
        with metadata that always contains the original content under the
        key defined in :data:`CONTENT`.

        Args:
            content: List of or singular chunks or strings to upsert.

        Returns:
            OneOrMany[str]: Generated ids for the inserted items.
        """
        ids = []
        embeddings = []
        metadatas = []
        documents = []
        is_many = True
        if isinstance(content, str):
            content = [content]
            is_many = False

        if isinstance(content, Chunk):
            content = [content]
            is_many = False

        for item in content:
            if isinstance(item, Chunk):
                id = item.id
                embedding = self._embedding_function([item.content])[0]
                metadata = item.metadata
                metadata[CONTENT] = item.content
                documents.append(item.document)

            else:
                id = str(uuid4())
                embedding = self._embedding_function([item])[0]
                metadata = {CONTENT: item}
                documents.append(None)

            ids.append(id)
            embeddings.append(embedding)
            metadatas.append(metadata)

        self._collection.upsert(
            ids=ids, embeddings=embeddings, metadatas=metadatas, documents=documents
        )
        return ids if is_many else ids[0]

    def fetch(
        self,
        ids: Optional[OneOrMany[str]] = None,
        where: Optional[Where] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        where_document: Optional[WhereDocument] = None,
    ) -> FetchResponse:
        """Fetch a set of vectors and their metadata from the collection.

        Args:
            ids: Optional list of ids or singular id to fetch.
            where: Optional metadata filter.
            limit: Result limit for pagination.
            offset: Result offset for pagination.
            where_document: Optional document-based filter.

        Returns:
            FetchResponse: A list-like container of :class:`FetchResult`.

        Raises:
            ValueError: If the Chroma response does not contain required fields.
        """
        results = FetchResponse()
        # currently we ignore Include and assume the default
        responses = self._collection.get(
            ids,
            where,
            limit,
            offset,
            where_document,
            include=["embeddings", "metadatas", "documents"],
        )

        embeddings = responses.get("embeddings")
        if embeddings is None:
            raise ValueError("Embeddings were not found in fetch response.")
        documents = responses.get("documents")
        if documents is None:
            raise ValueError("Documents were not found in fetch response.")
        metadatas = responses.get("metadatas")
        if metadatas is None:
            raise ValueError("Metadatas were not found in fetch response.")

        for i, response in enumerate(responses["ids"]):
            id = response

            metadata = dict(deepcopy(metadatas[i]))
            if not (content := metadata.get(CONTENT)) or not isinstance(content, str):
                raise ValueError(
                    "Content was not initialized in vector. Please create an issue"
                )

            metadata.pop(CONTENT)
            results.append(
                FetchResult(
                    id=id,
                    content=content,
                    vector=list(embeddings[i]),
                    document=documents[i],
                    metadata=metadata,
                )
            )

        return results

    # There is support for other types of query modalities but for now just list of strings
    # Should Probably add support for Chunks as well
    @overload
    def search(
        self,
        query: Chunk | str,
        ids: Optional[str] = None,
        top_k: int = 10,
        where: Optional[Where] = None,
        where_document: Optional[WhereDocument] = None,
        include: Include = [
            "metadatas",
            "embeddings",
            "documents",
            "distances",
        ],
    ) -> SearchResponse: ...

    @overload
    def search(
        self,
        query: list[Chunk] | list[str],
        ids: Optional[list[str]] = None,
        top_k: int = 10,
        where: Optional[Where] = None,
        where_document: Optional[WhereDocument] = None,
        include: Include = [
            "metadatas",
            "embeddings",
            "documents",
            "distances",
        ],
    ) -> list[SearchResponse]: ...

    def search(  # noqa: C901
        self,
        query: OneOrMany[Chunk] | OneOrMany[str],
        ids: Optional[OneOrMany[str]] = None,
        top_k: int = 10,
        where: Optional[Where] = None,
        where_document: Optional[WhereDocument] = None,
        include: Include = [
            "metadatas",
            "embeddings",
            "documents",
            "distances",
        ],
    ) -> OneOrMany[SearchResponse]:
        """Run a similarity search for the provided query texts.

        Args:
            query: A list of query strings or singular string to search for.
            ids: Optional list of ids or singular id to restrict the search to.
            top_k: Number of hits to return per query.
            where: Optional metadata filter to apply.
            where_document: Optional document filter to apply.
            include: Fields to include in the Chroma response.

        Returns:
            A list of :class:`SearchResponse` objects (one per query).

        Raises:
            ValueError: If expected fields are missing from the Chroma response.
        """
        is_many = True
        # If a single chunk is passed in, convert to list of string
        if isinstance(query, Chunk):
            query = [query.content]
            is_many = False

        # If a single string is passed in, convert to list of string
        elif isinstance(query, str):
            query = [query]
            is_many = False

        # If list of chunks is passed in, convert to list of strings
        elif isinstance(query, list) and all(isinstance(q, Chunk) for q in query):
            query = [q.content for q in query]

        elif isinstance(query, list) and all(isinstance(q, str) for q in query):
            pass
        else:
            raise ValueError(
                "Query must be a string, Chunk, or list of strings/Chunks."
            )

        query_embeddings = self._embedding_function(query)
        results = self._collection.query(
            query_embeddings=list(query_embeddings),
            ids=ids,
            n_results=top_k,
            where=where,
            where_document=where_document,
            include=include,
        )
        answer: list[SearchResponse] = []
        for query_idx, query_response in enumerate(results["ids"]):
            search_response = SearchResponse()
            for id_idx, id in enumerate(query_response):
                if not (distance := results.get("distances")):
                    raise ValueError("Distance not found in search results.")
                elif not (vector := results.get("embeddings")):
                    raise ValueError("Vector not found in search results.")
                elif not (document := results.get("documents")):
                    raise ValueError("Document not found in search results.")
                elif not (metadatas := results.get("metadatas")):
                    raise ValueError("Metadata not found in search results.")

                distance = distance[query_idx][id_idx]
                vector = list(vector[query_idx][id_idx])
                document = document[query_idx][id_idx]
                metadata = dict(deepcopy(metadatas[query_idx][id_idx]))

                if not (content := metadata.get(CONTENT)) or not isinstance(
                    content, str
                ):
                    raise ValueError(
                        "Content was not initialized in vector. Please create an issue"
                    )

                metadata.pop(CONTENT)

                search_response.append(
                    SearchResult(
                        id=id,
                        distance=distance,
                        content=content,
                        vector=vector,
                        document=document,  # Chroma document is just a str
                        metadata=metadata,
                    )
                )
            answer.append(search_response)

        return answer if is_many else answer[0]

    def delete(
        self,
        ids: OneOrMany[str],
        where: Optional[Where] = None,
        where_document: Optional[WhereDocument] = None,
    ):
        """
        Remove vectors from the store by id or metadata filter.
        Args:
            ids: list of ids or singular id to delete.
            where: Optional metadata filter.
            where_document: Optional document-based filter.
        """

        self._collection.delete(
            ids=ids,
            where=where,
            where_document=where_document,
        )

    def count(self) -> int:
        """Return the total number of vectors stored in the collection."""
        return self._collection.count()



# text_object.py
import hashlib
import logging
import os
from typing import Any, List, Optional

logger = logging.getLogger(__name__)


class ResourceInstance:
    """Base class for any resource object (text, image, video, etc.)."""

    def __init__(
        self,
        path: Optional[str] = None,
        name: Optional[str] = None,
        type: Optional[str] = None,
        description: Optional[str] = "",
        tags: Optional[List[str]] = None,
        comment: Optional[str] = "",
        hash: Optional[str] = None,
        duration: Optional[float] = None,
        object: Any = None,
    ):
        self.path = self.normalize_path(path) if path else None
        self.name = name or (self.get_name_from_path(self.path) if self.path else None)
        self.type = type or "unknown"
        self.description = description or ""
        self.tags = tags or []
        self.comment = comment or ""
        self.hash = hash or (self.get_resource_hash(self.path) if self.path else "")
        self.duration = duration or 0
        self.object = object

    @staticmethod
    def normalize_path(path: str) -> str:
        return path.replace("\\", "/")

    @staticmethod
    def get_name_from_path(path: str) -> str:
        name = os.path.basename(path)
        return os.path.splitext(name)[0]

    @staticmethod
    def get_resource_hash(file_content: str, _hash_type="sha256") -> str:
        hash_obj = hashlib.new(_hash_type)
        # hash on first 8192 bytes to avoid very large files
        hash_obj.update(file_content[:8192].encode("utf-8"))

        return hash_obj.hexdigest()

    def get_metadata(self) -> dict:
        """Serializable metadata dictionary."""
        return {
            "name": self.name,
            "type": self.type,
            "description": self.description,
            "path": self.path,
            "tags": self.tags,
            "comment": self.comment,
            "hash": self.hash,
            "duration": self.duration,
        }

    def __str__(self):
        return f"<ResourceInstance name={self.name}, type={self.type}>"


class TextObject(ResourceInstance):
    """Text-specific metadata and embedding storage."""

    def __init__(self, raw_content: str, path: Optional[str] = None, **kwargs):
        super().__init__(path=path, type="text", **kwargs)
        self.raw_content: str = raw_content
        self.hash: str = self.get_resource_hash(raw_content)
        self.chunked_content: List[str] = []
        self.embeddings: List[List[float]] = []

    def set_chunked(self, chunks: List[str]):
        self.chunked_content = chunks

    def set_embeddings(self, vectors: List[List[float]]):
        self.embeddings = vectors

    def get_metadata(self) -> dict:
        meta = super().get_metadata()
        meta.update(
            {
                "raw_content": self.raw_content,
                "num_chunks": len(self.chunked_content),
                "embeddings": f"{len(self.embeddings)} vectors",
            }
        )
        return meta



from __future__ import annotations

import uuid
from dataclasses import dataclass
from functools import reduce
from typing import Any, Dict, Iterable, List, Optional, Tuple

from railtracks.utils.profiling import Stamp
from railtracks.utils.serialization.graph import Edge

from .forest import (
    AbstractLinkedObject,
    Forest,
)


class Cancelled:
    pass


class Failure:
    def __init__(self, exception: Exception):
        self.exception = exception


@dataclass(frozen=True)
class RequestTemplate(AbstractLinkedObject):
    """
    A simple object containing details about a request in the system.
    """

    source_id: Optional[str]
    sink_id: str
    input: Tuple[Tuple, Dict]
    output: Optional[Any]
    parent: Optional[RequestTemplate]

    def __repr__(self):
        return f"RequestTemplate({self.identifier}, {self.source_id}, {self.sink_id}, {self.output}, {self.stamp})"

    def to_edge(self):
        """
        Converts the request template to an edge representation.
        """
        return Edge(
            source=self.source_id,
            target=self.sink_id,
            identifier=self.identifier,
            stamp=self.stamp,
            details={
                "input_args": self.input[0],
                "input_kwargs": self.input[1],
                "status": self.status,
                "output": self.output,
            },
            parent=self.parent.to_edge() if self.parent is not None else None,
        )

    @property
    def closed(self):
        """
        If the request has an output it is closed
        """
        return self.output is not None

    @property
    def is_insertion(self):
        """
        If the request is an insertion request it will return True.
        """
        return self.source_id is not None

    @property
    def status(self):
        """
        Gets the current status of the request.
        """

        if self.output is not None:
            if isinstance(self.output, Failure):
                return "Failed"

            return "Completed"
        else:
            return "Open"

    def get_all_parents(self) -> List[RequestTemplate]:
        """
        Recursely collects all the parents for the request.
        """
        if self.parent is None:
            return [self]
        return [self] + self.parent.get_all_parents()

    @property
    def get_terminal_parent(self):
        """
        Returns the terminal parent of the request.

        If this request is the parent then it will return itself.
        """
        if self.parent is not None:
            return self.parent.get_terminal_parent
        return self

    @property
    def duration_detail(self):
        """
        Returns the difference in time between the parent and the current request stamped time.
        """
        total_time = self.stamp.time - self.get_terminal_parent.stamp.time

        return total_time

    @classmethod
    def generate_id(cls):
        """
        Generates a new unique identifier for the request. This is suitable for use as a request identifier.
        """
        return str(uuid.uuid4())

    @classmethod
    def downstream(cls, requests: Iterable[RequestTemplate], source_id: Optional[str]):
        """Collects the requests one level downstream from the provided source_id."""
        return [x for x in requests if x.source_id == source_id]

    @classmethod
    def upstream(cls, requests: Iterable[RequestTemplate], sink_id: str):
        """Collects the requests one level upstream from the provided sink_id."""
        return [x for x in requests if x.sink_id == sink_id]

    @classmethod
    def all_downstream(
        cls, requests: Iterable[RequestTemplate], source_id: Optional[str]
    ):
        """
        Collects all the downstream requests from the provided source_id.
        """
        downstream_requests = cls.downstream(requests, source_id)
        additions = []
        for x in downstream_requests:
            additions += cls.all_downstream(requests, x.sink_id)

        return downstream_requests + additions

    @classmethod
    def open_tails(cls, requests: Iterable[RequestTemplate], source_id: Optional[str]):
        """
        Traverses down the provided tree to find all the open tails.

        Open Tail: is defined as any node which currently holds an open request and does not have any open ones beneath
        it.
        """
        open_downstream_requests = [
            x for x in cls.downstream(requests, source_id) if not x.closed
        ]

        # BASE CASE
        if len(open_downstream_requests) == 0:
            open_upstreams = cls.upstream(requests, source_id)
            assert len(open_upstreams) <= 1, (
                f"There should only be one or 0 upstream request, instead there was {len(open_upstreams)}"
            )
            return [r for r in open_upstreams if not r.closed]

        # RECURSIVE CASE
        return reduce(
            lambda x, y: x + y,
            [cls.open_tails(requests, x.sink_id) for x in open_downstream_requests],
            [],
        )

    @classmethod
    def children_complete(
        cls, requests: Iterable[RequestTemplate], source_node_id: str
    ):
        """
        Checks if all the downstream requests of a given parent node are complete. If so returns True.
         Otherwise, returns False.
        """
        downstream_requests = cls.downstream(requests, source_node_id)

        return all(x.closed for x in downstream_requests)


class RequestForest(Forest[RequestTemplate]):
    def __init__(
        self,
        request_heap: Dict[str, RequestTemplate] | None = None,
    ):
        """
        Creates a new instance of a request heap with no objects present.
        """
        super().__init__(heap=request_heap)

    def __getitem__(self, item):
        with self._lock:
            try:
                return self._heap[item]
            except KeyError:
                print(
                    f"failed to collect a request {item in [x.identifier for x in self._full_data]}"
                )
                raise

    def to_edges(self):
        """
        Converts the current heap into a list of `Edge` objects.
        """
        edge_list = [request.to_edge() for request in self._heap.values()]

        return edge_list

    def create(
        self,
        identifier: str,
        source_id: Optional[str],
        sink_id: str,
        input_args: Tuple,
        input_kwargs: Dict,
        stamp: Stamp,
    ):
        """
        Creates a new instance of a request from the provided details and places it into the heap.

        Args:
            identifier (str): The identifier of the request
            source_id (Optional[str]): The node id of the source, None if it is an insertion request.
            sink_id (str): The node id in the sink.
            input_args (Tuple): The input arguments for the request
            input_kwargs (Dict): The input keyword arguments for the request
            stamp (Stamp): The stamp that you would like this request to be tied to.

        Returns:
            str: The identifier of the request that was created.

        """
        # note we just need t be careful of any sort of race condition so we will be extra safe with our locking mechanism.
        with self._lock:
            if identifier in self._heap:
                raise RequestAlreadyExistsError(
                    f"You are trying to create a request {identifier} which already exists. "
                )

            new_request = RequestTemplate(
                identifier=identifier,
                source_id=source_id,
                sink_id=sink_id,
                input=(input_args, input_kwargs),
                output=None,
                stamp=stamp,
                parent=None,
            )

            self._update_heap(new_request)
            return identifier

    def update(
        self,
        identifier: str,
        output: Optional[Any],
        stamp: Stamp,
    ) -> None:
        """
        Updates the heap with the provided request details. Note you must call this function on a request that exist in the heap.

        The function will replace the old request with a new updated one with the provided output attached to the provided stamp.

        I will outline the special cases for this function:
        1. If you have provided a request id that does not exist in the heap, it will raise `RequestDoesNotExistError`

        Args:
            identifier (str): The identifier of the request
            output (Optional[RequestOutput]): The output of the request, None if the request is not completed.
            stamp (Stamp): The stamp that you would like this request addition to be tied to.

        Raises:
            RequestDoesNotExistError: If the request with the provided identifier does not exist in the heap.

        """
        with self._lock:
            old_request = self._heap.get(identifier, None)

            # one cannot update a request that does not exist
            if old_request is None:
                raise RequestDoesNotExistError(
                    f"Request with identifier {identifier} does not exist in the heap."
                )

            linked_request = RequestTemplate(
                identifier=identifier,
                source_id=old_request.source_id,
                sink_id=old_request.sink_id,
                input=old_request.input,
                output=output,
                stamp=stamp,
                parent=self._heap.get(identifier, None),
            )
            self._update_heap(linked_request)

    def children(self, parent_id: str):
        """
        Finds all the children of the provided parent_id.
        """
        with self._lock:
            return RequestTemplate.downstream(self._heap.values(), parent_id)

    @classmethod
    def generate_graph(
        cls, heap: Dict[str, RequestTemplate]
    ) -> Dict[str, List[Tuple[str, str]]]:
        """
        Generates a dictionary representation contain the edges in the graph. The key of the dictionary is the source
        and the value is a list of tuples where the first element is the sink_id and the second element is the request\
        id.
        Complexity: O(n) where n is the number of identifiers in the heap.

        Args:
            heap (Dict[str, RequestTemplate]): The heap of requests to generate the graph from.

        Returns:
            Dict[str, List[Tuple[str, str]]]: The graph representation of the heap.

        """

        # graph structure includes the request id and the source and sink id.
        # the tuple will be the sink id followed by the request id.
        graph: Dict[str, List[Tuple[str, str]]] = {}
        for request_temp in heap.values():
            if request_temp.source_id not in graph:
                graph[request_temp.source_id] = []
            if request_temp.sink_id not in graph:
                graph[request_temp.sink_id] = []

            graph[request_temp.source_id].append(
                (request_temp.sink_id, request_temp.identifier)
            )

        return graph

    def get_request_from_child_id(self, child_id: str):
        """
        Gets the request where this child_id is the sink_id of the request.

        Via the invariants of the system. There must only be 1 request that satisfies the above requirement.
        """
        with self._lock:
            upstreams = RequestTemplate.upstream(list(self._heap.values()), child_id)
            if len(upstreams) == 0:
                raise RequestDoesNotExistError(
                    f"Request with child id {child_id} does not exist in the heap."
                )
            assert len(upstreams) == 1, (
                f"Expected 1 upstream request, instead got {len(upstreams)}"
            )
            return upstreams[0]

    def open_tails(self):
        """
        Collects the current open tails in the heap. See `RequestTemplate.open_tails` for more information.
        """
        # the insertion request will have a none source_id
        with self._lock:
            o_t = RequestTemplate.open_tails(list(self._heap.values()), None)
            return o_t

    def children_requests_complete(self, parent_node_id: str):
        """
        Checks if all the downstream requests (one level down) if the given parent node are complete. If they are
         then it will return the request id of the parent node. Otherwise, it will return None.

        Note that you are providing the node_id of the parent node and downstream requests of that node is defined
         as any of the requests which have the matching parent_node.

        Args:
            parent_node_id (str): The parent node id

        Returns:
            The request_id string of the parent node if all the children are complete otherwise None.
        """

        with self._lock:
            if RequestTemplate.children_complete(
                list(self._heap.values()), parent_node_id
            ):
                upstreams = RequestTemplate.upstream(
                    list(self._heap.values()), parent_node_id
                )
                assert len(upstreams) == 1, (
                    f"Expected 1 upstream request, instead got {len(upstreams)}"
                )
                return upstreams[0].identifier
            return None

    @property
    def insertion_request(self):
        """
        Collects a list of all the insertion requests in the heap.

        They will be returned in the order that they were created.
        """
        insertions = [v for v in self._heap.values() if v.source_id is None]

        return insertions

    @property
    def answer(self):
        """
        Collects the answer to the insertion request.

        The behavior of the function can be split into two cases:

        1. There is either 1 or 0 insertion requests present:
         - In this case, it will return the output of the insertion request if it exists, otherwise None

        2. There is more than 1 insertion request:
         - Returns an ordered list of outputs of all the insertion requests. If one has not yet completed, it will
           return None in that index.
        """
        # first we must find the insertion request

        if len(self.insertion_request) == 0:
            return None
        if len(self.insertion_request) > 1:
            return [x.output for x in self.insertion_request]
        else:
            return self.insertion_request[0].output


class RequestDoesNotExistError(Exception):
    """
    A special exception to be thrown when you are trying to update a Request which does not exist.
    """

    pass


class RequestAlreadyExistsError(Exception):
    pass



"""
Tests for the docstring_parser module.

This module contains tests for the docstring parsing utilities in the
railtracks.llm.tools.docstring_parser module.
"""

from railtracks.llm.tools.docstring_parser import (
    parse_docstring_args,
    extract_args_section,
    parse_args_section,
    extract_main_description,
)


class TestExtractMainDescription:
    """Tests for the extract_main_description function."""

    def test_empty_docstring(self):
        """Test that an empty docstring returns an empty string."""
        assert extract_main_description("") == ""
        assert extract_main_description(None) == ""

    def test_simple_description(self):
        """Test extracting a simple description."""
        docstring = """This is a simple description."""
        assert extract_main_description(docstring) == "This is a simple description."

    def test_multiline_description(self):
        """Test extracting a multiline description."""
        docstring = """This is a multiline
        description that spans
        multiple lines."""
        expected = """This is a multiline
        description that spans
        multiple lines."""
        assert extract_main_description(docstring) == expected

    def test_description_with_sections(self):
        """Test extracting a description that has sections."""
        docstring = """This is the main description.

        Args:
            param1: Description of param1.
            
        Returns:
            The return value.
        """
        assert extract_main_description(docstring) == "This is the main description."


class TestExtractArgsSection:
    """Tests for the extract_args_section function."""

    def test_no_args_section(self):
        """Test a docstring without an Args section."""
        docstring = """This is a docstring without an Args section."""
        assert extract_args_section(docstring) == ""

    def test_simple_args_section(self):
        """Test extracting a simple Args section."""
        docstring = """
        This is a docstring.
        
        Args:
            param1: Description of param1.
            param2: Description of param2.
        
        Returns:
            The return value.
        """
        # The actual output includes the indentation and newlines
        result = extract_args_section(docstring)
        assert "param1: Description of param1." in result
        assert "param2: Description of param2." in result

    def test_args_section_with_types(self):
        """Test extracting an Args section with type annotations."""
        docstring = """
        This is a docstring.
        Args:
            param1 (str): Description of param1.
            param2 (int): Description of param2.
        """
        # The actual output includes the indentation and newlines
        result = extract_args_section(docstring)
        assert "param1 (str): Description of param1." in result
        assert "param2 (int): Description of param2." in result


class TestParseArgsSection:
    """Tests for the parse_args_section function."""

    def test_simple_args(self):
        """Test parsing simple Args."""
        args_section = """
            param1: Description of param1.
            param2: Description of param2.
        """
        expected = {
            "param1": "Description of param1.",
            "param2": "Description of param2.",
        }
        assert parse_args_section(args_section) == expected

    def test_args_with_types(self):
        """Test parsing Args with type annotations."""
        args_section = """
            param1 (str): Description of param1.
            param2 (int): Description of param2.
        """
        expected = {
            "param1": "Description of param1.",
            "param2": "Description of param2.",
        }
        assert parse_args_section(args_section) == expected

    def test_multiline_descriptions(self):
        """Test parsing Args with multiline descriptions."""
        args_section = """
            param1: Description of param1
                that spans multiple lines.
            param2: Description of param2.
        """
        expected = {
            "param1": "Description of param1 that spans multiple lines.",
            "param2": "Description of param2.",
        }
        assert parse_args_section(args_section) == expected


class TestParseDocstringArgs:
    """Tests for the parse_docstring_args function."""

    def test_empty_docstring(self):
        """Test parsing an empty docstring."""
        assert parse_docstring_args("") == {}
        assert parse_docstring_args(None) == {}

    def test_docstring_without_args(self):
        """Test parsing a docstring without Args section."""
        docstring = """This is a docstring without an Args section."""
        assert parse_docstring_args(docstring) == {}

    def test_simple_docstring(self):
        """Test parsing a simple docstring with Args."""
        docstring = """
        This is a docstring.
        
        Args:
            param1: Description of param1.
            param2: Description of param2.
        
        Returns:
            The return value.
        """
        expected = {
            "param1": "Description of param1.",
            "param2": "Description of param2.",
        }
        assert parse_docstring_args(docstring) == expected

    def test_complex_docstring(self):
        """Test parsing a complex docstring with various sections."""
        docstring = """
        This is a complex docstring.
        
        It has multiple paragraphs in the description.
        
        Args:
            param1 (str): Description of param1
                that spans multiple lines.
            param2 (int): Description of param2.
        
        Returns:
            The return value.
            
        Raises:
            ValueError: If something goes wrong.
        """
        expected = {
            "param1": "Description of param1 that spans multiple lines.",
            "param2": "Description of param2.",
        }
        assert parse_docstring_args(docstring) == expected

    def test_real_world_example(self):
        """Test parsing a real-world docstring example."""
        docstring = """
        Creates a new instance of a parameter object.

        Args:
            name: The name of the parameter.
            param_type: The type of the parameter.
            description: A description of the parameter.
            required: Whether the parameter is required. Defaults to True.
        """
        expected = {
            "name": "The name of the parameter.",
            "param_type": "The type of the parameter.",
            "description": "A description of the parameter.",
            "required": "Whether the parameter is required. Defaults to True.",
        }
        assert parse_docstring_args(docstring) == expected


class TestEdgeCases:
    """Tests for edge cases in docstring parsing."""

    def test_malformed_args_section(self):
        """Test parsing a malformed Args section."""
        docstring = """
        Args:
            This is not a proper parameter definition.
            param1: This is a proper one.
        """
        # Should only extract the properly formatted parameter
        expected = {"param1": "This is a proper one."}
        assert parse_docstring_args(docstring) == expected



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Optional
from uuid import uuid4

from .media_parser import MediaParser


@dataclass
class Chunk:
    """Structured chunk that can be upserted to a vector store.

    Attributes:
        content (str): The raw chunk text.
        id (Optional[str]): Identifier for the chunk. If not provided, a UUID
            is automatically generated in ``__post_init__``.
        document (Optional[str]): Optional document identifier or content
            associated with the chunk.
        metadata (dict[str, Any]): Arbitrary key-value metadata associated
            with this chunk. Defaults to an empty dictionary.

    """

    content: str
    id: Optional[str] = None
    document: Optional[str] = None
    metadata: Optional[dict[str, Any]] = field(default_factory=dict)

    def __post_init__(self) -> None:
        """Normalize metadata and ensure identifier is populated."""
        if self.metadata is None:
            self.metadata = {}
        if self.id is None:
            self.id = str(uuid4())


class BaseChunker(ABC):
    """Abstract base class for chunking strategies.

    A chunker splits input text ``Chunk`` objects. Specific chunking
    strategies should subclass this class and implement the abstract
    ``chunk`` method.

    Args:
        chunk_size (int): Specifies number of tokens per chunk to
            varying degree depending on implementation.
            Defaults to 400
        overlap (int): Specifies number of tokens to overlap
            between adjacent chunks to varying degree depending
            on implementation. Defaults to 200.

    Attributes:
        _chunk_size (int): Internal storage for chunk size.
        _overlap (int): Internal storage for overlap size.
    """

    def __init__(
        self,
        chunk_size: int = 400,
        overlap: int = 200,
    ):
        if not overlap < chunk_size:
            raise ValueError("'overlap' must be smaller than 'chunk_size'.")
        if not chunk_size > 0 or not overlap >= 0:
            raise ValueError(
                "'chunk_size' must be greater than 0 and 'overlap' must be at least 0 "
            )

        self._chunk_size = chunk_size
        self._overlap = overlap

    @property
    def chunk_size(self) -> int:
        return self._chunk_size

    @chunk_size.setter
    def chunk_size(self, value: int):
        if not self.overlap < value:
            raise ValueError("'overlap' must be smaller than 'chunk_size'.")
        if not value > 0:
            raise ValueError("'chunk_size' must be greater than 0")
        self._chunk_size = value

    @property
    def overlap(self) -> int:
        return self._overlap

    @overlap.setter
    def overlap(self, value: int):
        if not value < self._chunk_size:
            raise ValueError("'overlap' must be smaller than 'chunk_size'.")
        if not value >= 0:
            raise ValueError("'overlap' must be at least 0")
        self._overlap = value

    def chunk(
        self,
        text: str,
        document: Optional[str] = None,
        metadata: Optional[dict[str, Any]] = None,
    ) -> list[Chunk]:
        """Split text into list of chunks.


        Args:
            text (str): Raw text to chunk.
            document (Optional[str]): Identifier associated with the
                document or text source. Applied to each output chunk.
            metadata (dict[str, Any]): Additional metadata stored in each
                created chunk.

        Returns:
            list[Chunk]: A list of chunk objects produced by the chunking
            strategy.
        """
        text_chunk_list = self.split_text(text)
        chunks = self.make_into_chunks(text_chunk_list, document, metadata)

        return chunks

    def chunk_from_file(
        self,
        path: str,
        encoding: Optional[str] = None,
        document: Optional[str] = None,
        metadata: Optional[dict[str, Any]] = None,
    ) -> list[Chunk]:
        """Split text into list of chunks from a specified file.

        Args:
            path (str): File path to the input text source. Currently only `.pdf`
                and `.txt` files are supported.
            document (Optional[str]): Identifier associated with the
                document or text source. Applied to each created chunk.
            metadata (dict[str, Any]): Additional metadata stored in each
                created chunk.

        Returns:
            list[Chunk]: A list of chunk objects produced by the chunking
                strategy.

        """

        text = MediaParser.get_text(path, encoding=encoding)

        return self.chunk(text, document, metadata)

    @abstractmethod
    def split_text(
        self,
        text: str,
    ) -> list[str]:
        """Split text into text chunks.


        Args:
            text (str): Raw text to chunk.

        Returns:
            list[str]: A list of strings produced by the chunking strategy.
        """
        pass

    @classmethod
    def make_into_chunks(
        cls,
        text_chunks: list[str],
        document: Optional[str] = None,
        metadata: Optional[dict[str, Any]] = None,
    ) -> list[Chunk]:
        """Make list of text chunks into chunk type with associated attributes.


        Args:
            text_chunks (list[str]): Raw text to chunk.
            document (Optional[str]): Identifier associated with the
                document or text source. Applied to each output chunk.
            metadata (dict[str, Any]): Additional metadata stored in each
                created chunk.

        Returns:
            list[Chunk]: A list of chunk objects produced by the chunking
            strategy.

        """

        chunks = []
        for text_chunk in text_chunks:
            chunks.append(
                Chunk(
                    content=text_chunk,
                    document=document,
                    metadata=dict(metadata) if metadata is not None else None,
                )
            )
        return chunks



import asyncio
import time

import pytest
import railtracks as rt


def timeout_node(t_: float):
    time.sleep(t_)
    return t_


async def timeout_node_async(t_: float):
    await asyncio.sleep(t_)
    return t_


TimeoutNode = rt.function_node(timeout_node)
TimeoutNodeAsync = rt.function_node(timeout_node_async)


async def top_level_async():
    uno = rt.call(TimeoutNodeAsync, 1)
    dos = rt.call(TimeoutNodeAsync, 2)
    tres = rt.call(TimeoutNodeAsync, 3)
    quatro = rt.call(TimeoutNodeAsync, 2)
    cinco = rt.call(TimeoutNodeAsync, 1)

    result = await asyncio.gather(uno, dos, tres, quatro, cinco)

    return result


async def top_level():
    uno = rt.call(TimeoutNode, 1)
    dos = rt.call(TimeoutNode, 2)
    tres = rt.call(TimeoutNode, 3)
    quatro = rt.call(TimeoutNode, 2)
    cinco = rt.call(TimeoutNode, 1)

    result = await asyncio.gather(uno, dos, tres, quatro, cinco)

    return result


TopLevelAsync = rt.function_node(top_level_async)
TopLevel = rt.function_node(top_level)


@pytest.mark.timeout(4)
@pytest.mark.asyncio
@pytest.mark.parametrize("node", [TopLevel, TopLevelAsync], ids=["sync", "async"])
async def test_async_style_parallel(node):
    with rt.Session():
        result = await rt.call(node)
        assert result == [1, 2, 3, 2, 1]



import inspect
from typing import List

from mcp.server.fastmcp import FastMCP
from mcp.server.fastmcp.tools import Tool as MCPTool
from mcp.server.fastmcp.utilities.func_metadata import func_metadata

from railtracks.built_nodes.concrete import (
    RTFunction,
)

# TODO: this logic must be get refactored
from railtracks.interaction._call import call

# TODO this must be moved to a more appropriate place in utils.
from railtracks.llm.models._litellm_wrapper import _parameters_to_json_schema
from railtracks.nodes.nodes import Node


def _create_tool_function(
    node_cls: Node,
    node_info,
):
    type_map = {
        "integer": int,
        "number": float,
        "string": str,
        "boolean": bool,
        "array": list,
        "object": dict,
    }

    params = []
    args_doc = []
    params_schema = (
        _parameters_to_json_schema(node_info.parameters)
        if node_info.parameters is not None
        else {}
    )

    # Get all parameters and sort them: required first, then optional
    all_params = []
    required_params = set(params_schema.get("required", []))

    for param_name, param_info in params_schema.get("properties", {}).items():
        required = param_name in required_params
        param_type = param_info.get("type", "any")
        annotation = type_map.get(param_type, str)
        param_desc = param_info.get("description", "")

        all_params.append((param_name, param_info, required, annotation, param_desc))

    # Sort: required parameters first (True sorts before False when reversed)
    all_params.sort(key=lambda x: not x[2])  # not required = False sorts after True

    # Create parameters in the correct order
    for param_name, param_info, required, annotation, param_desc in all_params:
        if required:
            params.append(
                inspect.Parameter(
                    param_name,
                    inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    annotation=annotation,
                )
            )
        else:
            params.append(
                inspect.Parameter(
                    param_name,
                    inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    default=None,
                    annotation=annotation,
                )
            )

        args_doc.append(f"    {param_name}: {param_desc}")

    async def tool_function(**kwargs):
        return await call(node_cls.prepare_tool, **kwargs)

    tool_function.__signature__ = inspect.Signature(params)
    return tool_function


def create_mcp_server(
    nodes: List[Node | RTFunction],
    server_name: str = "MCP Server",
    fastmcp: FastMCP | None = None,
):
    """
    Create a FastMCP server that can be used to run nodes as MCP tools.

    Args:
        nodes: List of Node classes to be registered as tools with the MCP server.
        server_name: Name of the MCP server instance.
        fastmcp: Optional FastMCP instance to use instead of creating a new one.

    Returns:
        A FastMCP server instance.
    """
    if fastmcp is not None:
        if not isinstance(fastmcp, FastMCP):
            raise ValueError("Provided fastmcp must be an instance of FastMCP.")
        mcp = fastmcp
    else:
        mcp = FastMCP(server_name)

    for node in [n if not hasattr(n, "node_type") else n.node_type for n in nodes]:
        node_info = node.tool_info()
        func = _create_tool_function(node, node_info)

        mcp._tool_manager._tools[node_info.name] = MCPTool(
            fn=func,
            name=node_info.name,
            description=node_info.detail,
            parameters=(
                _parameters_to_json_schema(node_info.parameters)
                if node_info.parameters is not None
                else {}
            ),
            fn_metadata=func_metadata(func, []),
            is_async=True,
            context_kwarg=None,
            annotations=None,
        )  # Register the node as a tool

    return mcp



"""
Tests for the schema_parser module.

This module contains tests for the JSON output_schema parsing utilities in the
railtracks.llm.tools.schema_parser module.
"""

import re

from railtracks.llm.tools.schema_parser import (
    parse_json_schema_to_parameter,
    parse_model_properties,
)
from railtracks.llm.tools import Parameter, ArrayParameter, ObjectParameter, RefParameter


class TestParseJsonSchemaToParameter:
    """Tests for the parse_json_schema_to_parameter function."""

    def test_basic_string_parameter(self):
        """Test parsing a basic string parameter."""
        schema = {"type": "string", "description": "A string parameter"}
        param = parse_json_schema_to_parameter("test_param", schema, True)

        assert isinstance(param, Parameter)
        assert param.name == "test_param"
        assert param.param_type == "string"
        assert param.description == "A string parameter"
        assert param.required is True

    def test_basic_integer_parameter(self):
        """Test parsing a basic integer parameter."""
        schema = {"type": "integer", "description": "An integer parameter"}
        param = parse_json_schema_to_parameter("test_param", schema, False)

        assert isinstance(param, Parameter)
        assert param.name == "test_param"
        assert param.param_type == "integer"  # Should remain "integer"
        assert param.description == "An integer parameter"
        assert param.required is False

    def test_number_parameter_converts_to_float(self):
        """Test that 'number' type is converted to 'float'."""
        schema = {"type": "number", "description": "A number parameter"}
        param = parse_json_schema_to_parameter("test_param", schema, True)

        assert isinstance(param, Parameter)
        assert param.name == "test_param"
        assert param.param_type == "number"
        assert param.description == "A number parameter"
        assert param.required is True

    def test_boolean_parameter(self):
        """Test parsing a boolean parameter."""
        schema = {"type": "boolean", "description": "A boolean parameter"}
        param = parse_json_schema_to_parameter("test_param", schema, True)

        assert isinstance(param, Parameter)
        assert param.name == "test_param"
        assert param.param_type == "boolean"
        assert param.description == "A boolean parameter"
        assert param.required is True

    def test_array_parameter(self):
        """Test parsing an array parameter."""
        schema = {
            "type": "array",
            "description": "An array parameter",
            "items": {"type": "string"},
        }
        param = parse_json_schema_to_parameter("test_param", schema, True)

        assert isinstance(param, Parameter)
        assert param.name == "test_param"
        assert param.param_type == "array"
        assert param.description == "An array parameter"
        assert param.required is True

    def test_array_with_object_items(self):
        """Test parsing an array parameter with object items."""
        schema = {
            "type": "array",
            "description": "An array of objects",
            "items": {
                "type": "object",
                "properties": {
                    "name": {"type": "string", "description": "The name"},
                    "age": {"type": "integer", "description": "The age"},
                },
                "required": ["name"],
            },
        }
        param = parse_json_schema_to_parameter("test_param", schema, True)

        assert isinstance(param, ArrayParameter)
        assert param.name == "test_param"
        assert param.description == "An array of objects"
        assert param.required is True
        assert isinstance(param.items, ObjectParameter)
        # Check nested properties
        for prop in param.items.properties:
            assert isinstance(prop, Parameter), "Expected Parameter object"
            if prop.name == "name":
                assert prop.param_type == "string"
                assert prop.required is True
            elif prop.name == "age":
                assert prop.param_type == "integer"
                assert prop.required is False
            else:
                assert False, f"Unexpected property name: {prop.name}"

    def test_object_parameter(self):
        """Test parsing an object parameter."""
        schema = {
            "type": "object",
            "description": "An object parameter",
            "properties": {
                "name": {"type": "string", "description": "The name"},
                "age": {"type": "integer", "description": "The age"},
            },
            "required": ["name"],
        }
        param = parse_json_schema_to_parameter("test_param", schema, True)

        assert isinstance(param, ObjectParameter)
        assert param.name == "test_param"
        assert param.param_type == "object"
        assert param.description == "An object parameter"
        assert param.required is True

        # Check nested properties
        for prop in param.properties:
            assert isinstance(prop, Parameter), "Expected Parameter object"
            if prop.name == "name":
                assert prop.param_type == "string"
                assert prop.required is True
            elif prop.name == "age":
                assert prop.param_type == "integer"
                assert prop.required is False
            else:
                assert False, f"Unexpected property name: {prop.name}"

    def test_nested_object_parameter(self):
        """Test parsing a deeply nested object parameter."""
        schema = {
            "type": "object",
            "description": "A nested object parameter",
            "properties": {
                "person": {
                    "type": "object",
                    "description": "Person details",
                    "properties": {
                        "name": {"type": "string", "description": "The name"},
                        "address": {
                            "type": "object",
                            "description": "Address details",
                            "properties": {
                                "street": {
                                    "type": "string",
                                    "description": "Street name",
                                },
                                "city": {"type": "string", "description": "City name"},
                            },
                            "required": ["street"],
                        },
                    },
                    "required": ["name"],
                }
            },
            "required": ["person"],
        }
        param = parse_json_schema_to_parameter("test_param", schema, True)

        assert isinstance(param, ObjectParameter)
        assert param.name == "test_param"
        assert param.param_type == "object"

        # Convert param.properties set to a list for easier processing
        properties_list = list(param.properties)

        # Check first level nested property
        person_prop = next(p for p in properties_list if p.name == "person")
        assert person_prop.required is True
        assert person_prop.param_type == "object"

        # Check second level nested property
        person_properties_list = list(person_prop.properties)
        name_prop = next(p for p in person_properties_list if p.name == "name")
        assert name_prop.required is True

        address_prop = next(p for p in person_properties_list if p.name == "address")
        assert address_prop.param_type == "object"

        # Check third level nested property
        address_properties_list = list(address_prop.properties)
        street_prop = next(p for p in address_properties_list if p.name == "street")
        city_prop = next(p for p in address_properties_list if p.name == "city")

        assert street_prop.required is True
        assert city_prop.required is False

    def test_parameter_with_ref(self):
        """Test parsing a parameter with a $ref."""
        schema = {
            "$ref": "#/components/schemas/Person",
            "description": "A reference to Person output_schema",
        }
        param = parse_json_schema_to_parameter("test_param", schema, True)

        assert isinstance(param, RefParameter)
        assert param.name == "test_param"
        assert param.param_type == "object"
        assert param.description == "A reference to Person output_schema"
        assert param.required is True

    def test_default_type_is_object(self):
        """Test that the default type is 'object' when not specified."""
        schema = {"description": "A parameter without type"}
        param = parse_json_schema_to_parameter("test_param", schema, True)

        assert isinstance(param, Parameter)
        assert param.name == "test_param"
        assert param.param_type == "object"  # Default type
        assert param.description == "A parameter without type"
        assert param.required is True

    def test_empty_schema(self):
        """Test parsing an empty output_schema."""
        schema = {}
        param = parse_json_schema_to_parameter("test_param", schema, True)

        assert isinstance(param, Parameter)
        assert param.name == "test_param"
        assert param.param_type == "object"  # Default type
        assert param.description == ""
        assert param.required is True


class TestParseModelProperties:
    """Tests for the parse_model_properties function."""

    def test_simple_schema(self):
        """Test parsing a simple output_schema with basic properties."""
        schema = {
            "properties": {
                "name": {"type": "string", "description": "The name"},
                "age": {"type": "integer", "description": "The age"},
                "is_active": {
                    "type": "boolean",
                    "description": "Whether the user is active",
                },
            },
            "required": ["name", "age"],
        }

        result = parse_model_properties(schema)

        assert len(result) == 3
        assert all(isinstance(param, Parameter) for param in result)

        # Convert result set to a list for easier processing
        params_list = list(result)
        result = {param.name: param for param in params_list}

        assert "name" in result
        assert "age" in result
        assert "is_active" in result

        assert result["name"].param_type == "string"
        assert result["age"].param_type == "integer"
        assert result["is_active"].param_type == "boolean"

        assert result["name"].required is True
        assert result["age"].required is True
        assert result["is_active"].required is False

    def test_schema_with_nested_object(self):
        """Test parsing a output_schema with a nested object property."""
        schema = {
            "properties": {
                "name": {"type": "string", "description": "The name"},
                "address": {
                    "type": "object",
                    "description": "The address",
                    "properties": {
                        "street": {"type": "string", "description": "The street"},
                        "city": {"type": "string", "description": "The city"},
                    },
                    "required": ["street"],
                },
            },
            "required": ["name"],
        }

        result = parse_model_properties(schema)

        assert len(result) == 2
        assert all(isinstance(param, Parameter) for param in result)

        # Convert result set to a list for easier processing
        params_list = list(result)
        result = {param.name: param for param in params_list}

        assert "name" in result
        assert "address" in result

        # Check that address is a PydanticParameter with properties
        assert isinstance(result["address"], ObjectParameter)
        assert result["address"].param_type == "object"
        for param in result["address"].properties:
            assert isinstance(param, Parameter)
            if param.name == "street":
                assert param.required is True
            elif param.name == "city":
                assert param.required is False
            else:
                assert False, f"Unexpected property name: {param.name}"

    def test_schema_with_number_type(self):
        """Test parsing a output_schema with number type that should convert to float."""
        schema = {
            "properties": {"amount": {"type": "number", "description": "The amount"}}
        }

        result = parse_model_properties(schema)
        assert all(isinstance(param, Parameter) for param in result)
        params_list = list(result)
        result = {param.name: param for param in params_list}
        assert "amount" in result
        assert result["amount"].param_type == "number"

    def test_schema_with_defs_and_refs(self):
        """Test parsing a output_schema with $defs and $ref."""
        schema = {
            "$defs": {
                "Address": {
                    "properties": {
                        "street": {"type": "string", "description": "The street"},
                        "city": {"type": "string", "description": "The city"},
                    },
                    "required": ["street"],
                }
            },
            "properties": {
                "name": {"type": "string", "description": "The name"},
                "address": {"$ref": "#/$defs/Address", "description": "The address"},
            },
            "required": ["name"],
        }

        result = parse_model_properties(schema)

        assert len(result) == 2
        assert all(isinstance(param, Parameter) for param in result)
        params_list = list(result)
        result = {param.name: param for param in params_list}
        assert "name" in result
        assert "address" in result

        # Check that address is a PydanticParameter with properties from the $ref
        address_param = result["address"]
        assert isinstance(address_param, ObjectParameter)
        assert address_param.param_type == "object"
        address_properties = address_param.properties
        address_properties = {p.name: p for p in list(address_properties)}
        assert "street" in address_properties
        assert "city" in address_properties
        assert address_properties["street"].required is True
        assert address_properties["city"].required is False

    def test_schema_with_allof_and_refs(self):
        """Test parsing a output_schema with allOf and $ref."""
        schema = {
            "$defs": {
                "Person": {
                    "properties": {
                        "name": {"type": "string", "description": "The name"},
                        "age": {"type": "integer", "description": "The age"},
                    },
                    "required": ["name"],
                }
            },
            "properties": {
                "user": {
                    "allOf": [
                        {"$ref": "#/$defs/Person"},
                        {"type": "object", "description": "Additional user properties"},
                    ],
                    "description": "The user",
                }
            },
            "required": ["user"],
        }

        result = parse_model_properties(schema)

        assert len(result) == 1
        assert all(isinstance(param, Parameter) for param in result)
        params_list = list(result)
        result = {param.name: param for param in params_list}
        assert "user" in result

        # Check that user is a PydanticParameter with properties from the $ref
        assert isinstance(result["user"], ObjectParameter)
        assert result["user"].param_type == "object"
        sub_params = list(result["user"].properties)
        sub_params = {p.name: p for p in sub_params}
        assert "name" in sub_params
        assert "age" in sub_params
        assert sub_params["name"].required is True
        assert sub_params["age"].required is False

    def test_empty_schema(self):
        """Test parsing an empty output_schema."""
        schema = {}

        result = parse_model_properties(schema)

        assert result == []

    def test_schema_without_properties(self):
        """Test parsing a output_schema without properties."""
        schema = {
            "title": "Test Schema",
            "description": "A test output_schema without properties",
        }

        result = parse_model_properties(schema)

        assert result == []



import base64
import pytest
from pathlib import Path


from railtracks.llm.message import UserMessage
from railtracks.llm.encoding import ensure_data_uri

def make_png_base64():
    # Minimal PNG header + some bytes
    return base64.b64encode(b"\x89PNG\r\n\x1a\n123456").decode("utf-8")

# Supported extensions for the test folder
IMAGE_EXTENSIONS = [".png", ".jpg", ".jpeg", ".gif", ".webp", ".bmp", ".svg", ".ico", ".avif"]

def get_image_files(images_dir):
    # Use the absolute path from the repo root for safety
    repo_root = Path(__file__).parents[5]  # Adjust depth if needed
    images_path = repo_root / "packages" / "railtracks" / "tests" / "end_to_end" / "images"
    if not images_path.exists():
        return []
    return [
        f for f in images_path.iterdir()
        if f.is_file() and f.suffix.lower() in IMAGE_EXTENSIONS
    ]

@pytest.mark.parametrize("image_file", get_image_files("images"))
def test_user_message_with_local_file(image_file):
    msg = UserMessage(content="desc", attachment=str(image_file))
    att = msg.attachment[0]
    assert att.type == "local"
    assert att.encoding.startswith("data:image/")
    # Check that the encoding is valid base64
    header, b64 = att.encoding.split(",", 1)
    base64.b64decode(b64)

@pytest.mark.parametrize("image_file", get_image_files("images"))
def test_user_message_with_base64(image_file):
    with open(image_file, "rb") as f:
        b64 = base64.b64encode(f.read()).decode("utf-8")
    msg = UserMessage(content="desc", attachment=b64)
    att = msg.attachment[0]
    assert att.type == "data_uri"
    assert att.encoding.startswith("data:image/")
    header, b64_out = att.encoding.split(",", 1)
    base64.b64decode(b64_out)

@pytest.mark.parametrize("image_file", get_image_files("images"))
def test_user_message_with_data_uri(image_file):
    with open(image_file, "rb") as f:
        b64 = base64.b64encode(f.read()).decode("utf-8")
    ext = image_file.suffix.lower().lstrip(".")
    # Map extensions to mime types for test
    mime_map = {
        "jpg": "jpeg",
        "jpeg": "jpeg",
        "png": "png",
        "gif": "gif",
        "webp": "webp",
        "bmp": "bmp",
        "svg": "svg+xml",
        "ico": "x-icon",
        "avif": "avif",
    }
    mime = mime_map.get(ext, ext)
    data_uri = f"data:image/{mime};base64,{b64}"
    msg = UserMessage(content="desc", attachment=data_uri)
    att = msg.attachment[0]
    assert att.type == "data_uri"
    assert att.encoding == ensure_data_uri(data_uri)
    header, b64_out = att.encoding.split(",", 1)
    base64.b64decode(b64_out)

def test_user_message_with_plain_base64():
    b64 = make_png_base64()
    msg = UserMessage(content="desc", attachment=b64)
    assert msg.attachment is not None
    att = msg.attachment[0]
    assert att.type == "data_uri"
    assert att.encoding.startswith("data:image/png;base64,")

def test_user_message_with_valid_data_uri():
    b64 = make_png_base64()
    data_uri = f"data:image/png;base64,{b64}"
    msg = UserMessage(content="desc", attachment=data_uri)
    att = msg.attachment[0]
    assert att.type == "data_uri"
    assert att.encoding == ensure_data_uri(data_uri)

def test_user_message_with_invalid_data_uri_header():
    b64 = make_png_base64()
    bad_uri = f"data:image/png;base64{b64}"  # missing comma
    with pytest.raises(ValueError):
        UserMessage(content="desc", attachment=bad_uri)

def test_user_message_with_url(monkeypatch):
    # Patch urlopen to return PNG bytes
    import railtracks.llm.encoding as encoding_mod
    class DummyResponse:
        def __enter__(self): return self
        def __exit__(self, *a): return False
        def read(self): return b"\x89PNG\r\n\x1a\n123456"
    monkeypatch.setattr(encoding_mod.request, "urlopen", lambda url: DummyResponse())
    msg = UserMessage(content="desc", attachment="https://example.com/img.png")
    att = msg.attachment[0]
    assert att.type == "url"
    assert att.url == "https://example.com/img.png"

def test_user_message_multiple_attachments(tmp_path):
    img_file = tmp_path / "img.png"
    img_file.write_bytes(b"\x89PNG\r\n\x1a\n123456")
    b64 = make_png_base64()
    msg = UserMessage(content="desc", attachment=[str(img_file), b64])
    assert len(msg.attachment) == 2
    assert msg.attachment[0].type == "local"
    assert msg.attachment[1].type == "data_uri"


from railtracks.llm import MessageHistory, SystemMessage, UserMessage
from railtracks.llm.prompt_injection_utils import ValueDict


def inject_values(message_history: MessageHistory, value_dict: ValueDict):
    """
    Injects the values in the `value_dict` from the current request into the prompt.

    Args:
        message_history (MessageHistory): The prompts to inject context into.
        value_dict (ValueDict): The dictionary containing values to fill in the prompt.

    """

    for message in message_history:
        if message.inject_prompt and isinstance(message.content, str):
            try:
                if isinstance(message, (UserMessage, SystemMessage)):
                    message.fill_prompt(value_dict)
                    message.inject_prompt = False
            except ValueError:
                pass

    return message_history



from __future__ import annotations

from abc import ABC
from typing import Any, Literal, ParamSpec, Type, TypeVar

from railtracks.nodes.nodes import Node, NodeState

# RT specific imports

ExecutionConfigurations = Literal["async"]

_P = ParamSpec("_P")
_TOutput = TypeVar("_TOutput")
_TNode = TypeVar("_TNode", bound=Node)

############### Request Completion Messages ###############


class RequestCompletionMessage(ABC):
    """
    The base class for all messages on the request completion system.
    """

    def log_message(self) -> str:
        """Converts the message to a string ready to be logged."""
        return repr(self)


### Request Finished Messages ########
# TODO add generic typing for all these types


class RequestFinishedBase(RequestCompletionMessage, ABC):
    """
    A simple base class for all messages that pertain to a request finishing.
    """

    def __init__(
        self,
        *,
        request_id: str,
        node_state: NodeState[Node[_TOutput]] | None,
    ):
        self.request_id = request_id
        self.node_state = node_state

    @property
    def node(self) -> Node[_TOutput] | None:
        """
        Gets a node instance from the provided node state.

        Note: this function uses the functionality of `NodeState.instantiate()`
        """
        if self.node_state is None:
            return None

        return self.node_state.instantiate()

    def __repr__(self):
        return f"{self.__class__.__name__}(request_id={self.request_id}, node_state={self.node_state})"


class RequestSuccess(RequestFinishedBase):
    """
    A message that indicates the succseful completion of a request.
    """

    def __init__(
        self,
        *,
        request_id: str,
        node_state: NodeState[_TNode[_TOutput]],
        result: _TOutput,
    ):
        super().__init__(request_id=request_id, node_state=node_state)
        self.result = result

    def __repr__(self):
        return f"{self.__class__.__name__}(request_id={self.request_id}, node_state={self.node_state}, result={self.result})"

    def log_message(self) -> str:
        return f"{self.node_state.node.name()} DONE with result {self.result}"


class RequestFailure(RequestFinishedBase):
    """
    A message that indicates a failure in the request execution.
    """

    def __init__(
        self,
        *,
        request_id: str,
        node_state: NodeState[_TNode[_TOutput]] | None,
        error: Exception,
    ):
        super().__init__(request_id=request_id, node_state=node_state)
        self.error = error

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(request_id={self.request_id}, "
            f"node_state={self.node_state}, error={self.error})"
        )

    def log_message(self) -> str:
        return f"{self.node_state.node.name()} FAILED with error {self.error}"


class RequestCreationFailure(RequestFinishedBase):
    """
    A special class for situations where the creation of a new request fails before it was ever able to run.
    """

    def __init__(self, *, request_id: str, error: Exception):
        super().__init__(request_id=request_id, node_state=None)
        self.error = error

    def __repr__(self):
        return f"{self.__class__.__name__}(request_id={self.request_id}, error={self.error})"

    def log_message(self) -> str:
        return f"Request creation FAILED with error {self.error}"


####### Request Creation Messages ########


class RequestCreation(RequestCompletionMessage):
    """
    A message that describes the creation of a new request in the system.
    """

    def __init__(
        self,
        *,
        current_node_id: str | None,
        current_run_id: str | None,
        new_request_id: str,
        running_mode: ExecutionConfigurations,
        new_node_type: Type[Node],
        args,
        kwargs,
    ):
        self.current_node_id = current_node_id
        self.current_run_id = current_run_id
        self.new_request_id = new_request_id
        self.running_mode = running_mode
        self.new_node_type = new_node_type
        self.args = args
        self.kwargs = kwargs

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(current_node_id={self.current_node_id}, "
            f"new_request_id={self.new_request_id}, running_mode={self.running_mode}, "
            f"new_node_type={self.new_node_type.__name__}, args={self.args}, kwargs={self.kwargs})"
        )


##### OTHER MESSAGES #####


class FatalFailure(RequestCompletionMessage):
    """
    A message that indicates an irrecoverable failure in the request completion system.
    """

    def __init__(self, *, error: Exception):
        self.error = error

    def __repr__(self):
        return f"{self.__class__.__name__}(error={self.error})"


class Streaming(RequestCompletionMessage):
    """
    A message that indicates a streaming operation in the request completion system.
    """

    def __init__(self, *, streamed_object: Any, node_id: str):
        self.streamed_object = streamed_object
        self.node_id = node_id

    def __repr__(self):
        return f"{self.__class__.__name__}(streamed_object={self.streamed_object}, node_id={self.node_id})"



from __future__ import annotations

import asyncio
import uuid
from typing import Callable, Coroutine, Generic, List, TypeVar

from .logging.create import get_rt_logger

_T = TypeVar("_T")
_TOutput = TypeVar("_TOutput")

logger = get_rt_logger("Publisher")


class Subscriber(Generic[_T]):
    """A simple wrapper class of a callback function."""

    # this could be done without a class, but I want to keep as extendable as possible.
    def __init__(
        self,
        callback: Callable[[_T], None] | Callable[[_T], Coroutine[None, None, None]],
        name: str | None = None,
    ):
        self.callback = callback
        self.name = name if name is not None else callback.__name__
        self.id = str(uuid.uuid4())

    async def trigger(self, message: _T):
        """Trigger this broadcast_callback with the given message."""
        try:
            result = self.callback(message)
            if asyncio.iscoroutine(result):
                await result
        except Exception as e:
            logger.debug(msg=f"Error in {self.name}", exc_info=e)


class Publisher(Generic[_T]):
    """
    A simple publisher object with some basic functionality to publish and suvbscribe to messages.

    Note a couple of things:
    - Message will be handled in the order they came in (no jumping the line)
    - If you add a broadcast_callback during the operation it will handle any new messages that come in after the subscription
        took place
    - Calling the shutdown method will kill the publisher forever. You will have to make a new one after.
    """

    timeout = 0.001

    def __init__(
        self,
    ):
        self._queue: asyncio.Queue[_T] | None = None
        self._subscribers: List[Subscriber[_T]] = []

        self._running = False

        self.pub_loop = None

    async def __aenter__(self):
        """Enable the use of the publisher in a context manager."""
        await self.start()
        return self

    async def start(self):
        # you must set the kill variables first or the publisher loop will early exit.
        self._running = True
        self._queue = asyncio.Queue()
        self.pub_loop = asyncio.create_task(
            self._published_data_loop(), name="Publisher Loop"
        )

    async def __aexit__(self, exc_type, exc_value, traceback):
        """Shutdown the publisher when exiting the context manager."""
        await self.shutdown()

    async def publish(self, message: _T):
        """Publish a message the publisher. This will trigger all subscribers to receive the message.

        Args:
            message: The message you would like to publish.

        """
        # logger.debug(f"Publishing message: {message}")
        if not self._running:
            raise RuntimeError("Publisher is not currently running.")

        await self._queue.put(message)

    async def _published_data_loop(self):
        """
        A loop designed to be run in a thread that will continuously check for new messages in the queue and trigger
        subscribers as they are received
        """
        while self._running:
            try:
                message = await asyncio.wait_for(
                    self._queue.get(), timeout=self.timeout
                )

                try:
                    contracts = [sub.trigger(message) for sub in self._subscribers]

                    await asyncio.gather(*contracts)

                # we need a broad exception clause to catch any errors that might occur in the subs.
                except Exception:
                    pass

                # will only reach this section after all the messages have been handled

            # this exception is raised when the queue is empty for `self.timeout` seconds.
            # we do this so we can check is the self._running flag.
            except asyncio.TimeoutError:
                continue

    def subscribe(
        self,
        callback: Callable[[_T], None] | Callable[[_T], Coroutine[None, None, None]],
        name: str | None = None,
    ) -> str:
        """
        Subscribe the publisher so whenever we receive a message the callback will be triggered.

        Args:
            callback: The callback function that will be triggered when a message is published.
            name: Optional name for the broadcast_callback, mainly used for debugging.

        Returns:
            str: A unique identifier for the broadcast_callback. You can use this key to unsubscribe later.

        """
        sub = Subscriber(callback, name)
        self._subscribers.append(sub)
        return sub.id

    def unsubscribe(self, identifier: str):
        """
        Unsubscribe the publisher so the given broadcast_callback will no longer receive messages.

        Args:
            identifier: The unique identifier of the broadcast_callback to remove.

        Raises:
            KeyError: If no broadcast_callback with the given identifier exists.
        """
        index_to_remove = [
            index for index, sub in enumerate(self._subscribers) if sub.id == identifier
        ]
        if not index_to_remove:
            raise KeyError(f"No broadcast_callback with identifier {identifier} found.")

        index_to_remove = index_to_remove[0]
        self._subscribers.pop(index_to_remove)

    async def listener(
        self,
        message_filter: Callable[[_T], bool],
        result_mapping: Callable[[_T], _TOutput] = lambda x: x,
        listener_name: str | None = None,
    ):
        """
        Creates a special listener object that will wait for the first message that matches the given filter.

        After receiving the message it will run the result_mapping function on the message and return the result, and
        kill the broadcast_callback.

        Args:
            message_filter: A function that takes a message and returns True if the message should be returned.
            result_mapping: A function that maps the message into a final result.
            listener_name: Optional name for the listener, mainly used for debugging.
        """

        async def single_listener():
            returnable_result: asyncio.Future[_T] = asyncio.Future()
            # we are gonna use the asyncio.event system instead of threading
            listener_event = asyncio.Event()

            async def special_subscriber(message: _T):
                nonlocal returnable_result
                if message_filter(message):
                    # this will trigger the end of the listener loop
                    returnable_result.set_result(message)
                    listener_event.set()
                    return

            sub_id = self.subscribe(
                callback=special_subscriber,
                name=listener_name,
            )

            while True:
                try:
                    await asyncio.wait_for(listener_event.wait(), timeout=self.timeout)
                    break
                except asyncio.TimeoutError:
                    pass

                if not self._running:
                    raise ValueError(
                        "Listener has been killed before receiving the correct message."
                    )

            unwrapped_returned_result: _T = returnable_result.result()
            self.unsubscribe(sub_id)
            return result_mapping(unwrapped_returned_result)

        return await single_listener()

    async def shutdown(self):
        """
        Shutdowns the publisher and halts the listener loop.

        Note that this will work slowly, as it will wait for the current messages in the queue to be processed before
        shutting down.
        """

        self._running = False
        await self.pub_loop

        return

    def is_running(self) -> bool:
        """
        Check if the publisher is currently running.

        Returns:
            bool: True if the publisher is running, False otherwise.
        """
        return self._running



import os
import pytest
import requests
from unittest.mock import patch, MagicMock

import litellm

from railtracks.llm.models.local.ollama import OllamaLLM
from railtracks.llm._exception_base import RTLLMError
from railtracks.llm.history import MessageHistory
from railtracks.llm.message import UserMessage

@pytest.fixture
def mock_response():
    """Fixture for mocking successful API response"""
    mock = MagicMock()
    mock.json.return_value = {
        "models": [
            {"name": "test-model"},
            {"name": "llama2"}
        ]
    }
    mock.raise_for_status.return_value = None
    return mock


@pytest.fixture
def mock_failed_response():
    """Fixture for mocking failed API response"""
    mock = MagicMock()
    mock.raise_for_status.side_effect = requests.exceptions.RequestException("Connection failed")
    return mock


def test_model_type():
    """Test the model_type class method"""
    assert OllamaLLM.model_gateway() == "Ollama"
def test_init_success(mock_response):
    """Test successful initialization of Ollama"""
    with patch('requests.get', return_value=mock_response):
        ollama = OllamaLLM("test-model")
        assert ollama.model_name() == "ollama/test-model"
        assert ollama.domain == "http://localhost:11434"


def test_init_with_custom_domain(mock_response):
    """Test initialization with custom domain"""
    custom_domain = "http://custom-domain:11434"
    with patch('requests.get', return_value=mock_response):
        ollama = OllamaLLM("test-model", domain="custom", custom_domain=custom_domain)
        assert ollama.domain == custom_domain


def test_init_model_not_available(mock_response):
    """Test initialization with unavailable model"""
    with patch('requests.get', return_value=mock_response):
        with pytest.raises(RTLLMError) as exc_info:
            OllamaLLM("unavailable-model")
        assert "not available on server" in str(exc_info.value)


def test_init_connection_error(mock_failed_response):
    """Test initialization with connection error"""
    with patch('requests.get', return_value=mock_failed_response):
        with pytest.raises(requests.exceptions.RequestException):
            OllamaLLM("test-model")


def test_run_check_success(mock_response):
    """Test successful API check"""
    with patch('requests.get', return_value=mock_response):
        ollama = OllamaLLM("test-model")
        # Should not raise any exception
        ollama._run_check("api/tags")


def test_run_check_connection_error(mock_failed_response):
    """Test API check with connection error"""
    with patch('requests.get', side_effect=requests.exceptions.RequestException("Connection failed")):
        with pytest.raises(requests.exceptions.RequestException):
            OllamaLLM("test-model")


def test_chat_with_tools_unsupported(mock_response):
    """Test chat_with_tools with unsupported model"""
    with patch('requests.get', return_value=mock_response):
        with patch.object(litellm, 'supports_function_calling', return_value=False):
            ollama = OllamaLLM("test-model")
            messages = MessageHistory([UserMessage(content="test message")])
            tools = []

            with pytest.raises(RTLLMError) as exc_info:
                ollama.chat_with_tools(messages, tools)
            assert "does not support function calling" in str(exc_info.value)


def test_model_name_extraction(mock_response):
    """Test model name extraction from full path"""
    with patch('requests.get', return_value=mock_response):
        ollama = OllamaLLM("ollama/test-model")
        assert ollama.model_name() == "ollama/test-model"
        ollama2 = OllamaLLM("test-model")
        assert ollama2.model_name() == "ollama/test-model"


def test_init_with_auto_domain_missing_env(mock_response):
    """Test initialization with auto domain but missing environment variable"""
    with patch('requests.get', return_value=mock_response):
        with patch.dict(os.environ, {}, clear=True):
            with pytest.raises(RTLLMError) as exc_info:
                OllamaLLM("test-model", domain="auto")
            assert "OLLAMA_HOST environment variable not set" in str(exc_info.value)


def test_init_with_custom_domain_missing_env(mock_response):
    """Test initialization with custom domain but missing environment variable"""
    with patch('requests.get', return_value=mock_response):
        with patch.dict(os.environ, {}, clear=True):
            with pytest.raises(RTLLMError) as exc_info:
                OllamaLLM("test-model", domain="auto")
            assert "OLLAMA_HOST environment variable not set" in str(exc_info.value)


def test_init_with_default_domain(mock_response):
    """Test initialization with default domain setting"""
    with patch('requests.get', return_value=mock_response):
        ollama = OllamaLLM("test-model", domain="default")
        assert ollama.domain == "http://localhost:11434"


def test_init_with_auto_domain(mock_response):
    """Test initialization with auto domain"""
    custom_domain = "http://custom-domain:11434"
    with patch('requests.get', return_value=mock_response):
        with patch.dict(os.environ, {'OLLAMA_HOST': custom_domain}):
            ollama = OllamaLLM("test-model", domain="auto")
            assert ollama.domain == custom_domain


def test_init_with_custom_domain_missing_arg(mock_response):
    """Test initialization with custom domain but missing custom_domain argument"""
    with patch('requests.get', return_value=mock_response):
        with pytest.raises(RTLLMError) as exc_info:
            OllamaLLM("test-model", domain="custom")
        assert "Custom domain must be provided" in str(exc_info.value)



from .messages import (
    FatalFailure,
    RequestCompletionMessage,
    RequestCreation,
    RequestCreationFailure,
    RequestFailure,
    RequestFinishedBase,
    RequestSuccess,
    Streaming,
)

__all__ = [
    "RequestCompletionMessage",
    "RequestCreationFailure",
    "RequestFailure",
    "RequestCreation",
    "RequestSuccess",
    "RequestFinishedBase",
    "FatalFailure",
    "Streaming",
    "output_mapping",
    "RTPublisher",
    "stream_subscriber",
]

from ._subscriber import stream_subscriber
from .publisher import RTPublisher
from .utils import output_mapping



from .content import ToolCall, ToolResponse
from .history import MessageHistory
from .message import AssistantMessage, Message, SystemMessage, ToolMessage, UserMessage
from .model import ModelBase
from .models import (
    AnthropicLLM,
    AzureAILLM,
    CohereLLM,
    GeminiLLM,
    HuggingFaceLLM,
    OllamaLLM,
    OpenAICompatibleProvider,
    OpenAILLM,
    PortKeyLLM,
    # TelusLLM,
)
from .providers import ModelProvider
from .tools import (
    ArrayParameter,
    ObjectParameter,
    Parameter,
    RefParameter,
    Tool,
    UnionParameter,
)

__all__ = [
    "ModelBase",
    "ToolCall",
    "ToolResponse",
    "UserMessage",
    "SystemMessage",
    "AssistantMessage",
    "Message",
    "ToolMessage",
    "MessageHistory",
    "ModelProvider",
    "Tool",
    "AnthropicLLM",
    "AzureAILLM",
    "CohereLLM",
    "HuggingFaceLLM",
    "OpenAILLM",
    "GeminiLLM",
    "OllamaLLM",
    "AzureAILLM",
    "GeminiLLM",
    # "TelusLLM",
    "PortKeyLLM",
    "OpenAICompatibleProvider",
    "CohereLLM",
    # Parameter types
    "Parameter",
    "UnionParameter",
    "ArrayParameter",
    "ObjectParameter",
    "RefParameter",
]



import pytest
from unittest.mock import patch
from railtracks.pubsub.publisher import RTPublisher
# ================================== Pub Sub Fixtures ==================================
@pytest.fixture
def dummy_publisher():
    pub = RTPublisher()
    return pub

@pytest.fixture
def logger_patch():
    with patch("railtracks.utils.publisher.logger") as mock_logger:
        yield mock_logger

# ================================== Message Fixtures ==================================
@pytest.fixture
def dummy_node_class():
    class Node:
        def name(self):
            return "MockNode"
    return Node

@pytest.fixture
def dummy_node_state(dummy_node_class):
    class DummyNodeState:
        def __init__(self):
            self.node = dummy_node_class()
        def instantiate(self):
            return self.node
        def __repr__(self):
            return f"DummyNodeState({self.node})"
    return DummyNodeState()

@pytest.fixture
def dummy_exception():
    return Exception("Something went wrong")


from __future__ import annotations

import contextvars
import logging
import os
import warnings
from typing import TYPE_CHECKING, Any, Callable, Coroutine, KeysView

from railtracks.exceptions import ContextError

if TYPE_CHECKING:
    from railtracks.pubsub.publisher import RTPublisher

from railtracks.utils.config import ExecutorConfig
from railtracks.utils.logging.config import AllowableLogLevels

from ..utils.logging.config import configure_module_logging
from .external import ExternalContext, MutableExternalContext
from .internal import InternalContext


class RunnerContextVars:
    """
    A class to hold context variables which are scoped within the context of a single runner.
    """

    def __init__(
        self,
        *,
        internal_context: InternalContext,
        external_context: ExternalContext,
    ):
        self.internal_context = internal_context
        self.external_context = external_context

    def prepare_new(self, new_parent_id: str, new_run_id: str | None = None):
        """
        Update the parent ID of the internal context.
        """
        new_internal_context = self.internal_context.prepare_new(
            new_parent_id=new_parent_id,
            run_id=new_run_id,
        )

        return RunnerContextVars(
            internal_context=new_internal_context,
            external_context=self.external_context,
        )


runner_context: contextvars.ContextVar[RunnerContextVars | None] = (
    contextvars.ContextVar("runner_context", default=None)
)

global_executor_config: contextvars.ContextVar[ExecutorConfig] = contextvars.ContextVar(
    "executor_config", default=ExecutorConfig()
)


def safe_get_runner_context() -> RunnerContextVars:
    """
    Safely get the runner context for the current thread.

        Returns:
            RunnerContextVars: The runner context associated with the current thread.

        Raises:
            ContextError: If the global variables have not been registered.
    """
    context = runner_context.get()
    if context is None:
        raise ContextError(
            message="Context is not available. But some function tried to access it.",
            notes=[
                "You need to have an active runner to access context.",
                "Eg.-\n with rt.Session():\n    _ = rt.call(node)",
            ],
        )
    return context


def is_context_present():
    """Returns true if a context exists."""
    t_c = runner_context.get()
    return t_c is not None


def is_context_active():
    """
    Check if the global variables for the current thread are active.

    Returns:
        bool: True if the global variables are active, False otherwise.
    """
    context = runner_context.get()
    return context is not None and context.internal_context.is_active


def get_publisher() -> RTPublisher:
    """
    Get the publisher for the current thread's global variables.

    Returns:
        RTPublisher: The publisher associated with the current thread's global variables.

    Raises:
        RuntimeError: If the global variables have not been registered.
    """
    context = safe_get_runner_context()
    return context.internal_context.publisher


def get_session_id() -> str | None:
    """
    Get the runner ID of the current thread's global variables.

    Returns:
        str: The runner ID associated with the current thread's global variables.

    Raises:
        ContextError: If the global variables have not been registered.
    """
    context = safe_get_runner_context()
    return context.internal_context.session_id


def get_parent_id() -> str | None:
    """
    Get the parent ID of the current thread's global variables.

    Returns:
        str | None: The parent ID associated with the current thread's global variables, or None if not set.

    Raises:
        ContextError: If the global variables have not been registered.
    """
    context = safe_get_runner_context()
    return context.internal_context.parent_id


def get_run_id() -> str | None:
    """
    Get the run ID of the current thread's global variables.

    Returns:
        str | None: The run ID associated with the current thread's global variables, or None if not set.


    Raises:
        ContextError: If the global variables have not been registered.
    """
    context = safe_get_runner_context()
    return context.internal_context.run_id


def register_globals(
    *,
    session_id: str,
    rt_publisher: RTPublisher | None,
    parent_id: str | None,
    executor_config: ExecutorConfig,
    global_context_vars: dict[str, Any],
):
    """
    Register the global variables for the current thread.
    """
    i_c = InternalContext(
        publisher=rt_publisher,
        parent_id=parent_id,
        session_id=session_id,
        executor_config=executor_config,
    )
    e_c = MutableExternalContext(global_context_vars)

    runner_context_vars = RunnerContextVars(
        internal_context=i_c,
        external_context=e_c,
    )

    runner_context.set(runner_context_vars)


async def activate_publisher():
    """
    Activate the publisher for the current thread's global variables.

    This function should be called to ensure that the publisher is running and can be used to publish messages.
    """
    r_c = safe_get_runner_context()
    internal_context = r_c.internal_context
    assert internal_context is not None

    assert internal_context.publisher is not None

    await internal_context.publisher.start()


async def shutdown_publisher():
    """
    Shutdown the publisher for the current thread's global variables.

    This function should be called to stop the publisher and clean up resources.
    """
    context = safe_get_runner_context()
    context = context.internal_context
    assert context is not None

    assert context.publisher.is_running()
    await context.publisher.shutdown()


def get_global_config() -> ExecutorConfig:
    """
    Get the executor configuration for the current thread's global variables.

    Returns:
        ExecutorConfig: The executor configuration associated with the current thread's global variables, or None if not set.
    """
    executor_config = global_executor_config.get()
    return executor_config


def get_local_config() -> ExecutorConfig:
    """
    Get the executor configuration for the current thread's global variables.

    Returns:
        ExecutorConfig: The executor configuration associated with the current thread's global variables, or None if not set.
    """
    context = safe_get_runner_context()

    return context.internal_context.executor_config


def set_local_config(
    executor_config: ExecutorConfig,
):
    """
    Set the executor configuration for the current thread's global variables.

    Args:
        executor_config (ExecutorConfig): The executor configuration to set.
    """
    context = safe_get_runner_context()

    context.executor_config = executor_config
    runner_context.set(context)


def set_global_config(
    executor_config: ExecutorConfig,
):
    """
    Set the executor configuration for the current thread's global variables.

    Args:
        executor_config (ExecutorConfig): The executor configuration to set.
    """
    global_executor_config.set(executor_config)


def update_parent_id(new_parent_id: str, new_run_id: str | None = None):
    """
    Update the parent ID of the current thread's global variables.

    If no run ID is provided, the current run ID will be used.
    """
    current_context = safe_get_runner_context()

    assert (
        new_run_id is not None or current_context.internal_context.run_id is not None
    ), "You cannot update the parent ID while a run ID is inactive"

    if current_context is None:
        raise RuntimeError("No global variable set")

    new_context = current_context.prepare_new(new_parent_id, new_run_id=new_run_id)

    runner_context.set(new_context)


def delete_globals():
    """Resets the globals to None."""
    runner_context.set(None)


def get(
    key: str,
    /,
    default: Any | None = None,
):
    """
    Get a value from context

    Args:
        key (str): The key to retrieve.
        default (Any | None): The default value to return if the key does not exist. If set to None and the key does not exist, a KeyError will be raised.
    Returns:
        Any: The value associated with the key, or the default value if the key does not exist.

    Raises:
        KeyError: If the key does not exist and no default value is provided.
    """
    context = safe_get_runner_context()
    return context.external_context.get(key, default=default)


def put(
    key: str,
    value: Any,
):
    """
    Set a value in the context.

    Args:
        key (str): The key to set.
        value (Any): The value to set.
    """
    context = safe_get_runner_context()
    context.external_context.put(key, value)


def update(data: dict[str, Any]):
    """
    Sets the values in the context. If the context already has values, this will overwrite them, but it will not delete any existing keys.

    Args:
        data (dict[str, Any]): The data to update the context with.
    """
    context = safe_get_runner_context()
    context.external_context.update(data)


def delete(key: str):
    """
    Delete a key from the context.

    Args:
        key (str): The key to delete.

    Raises:
        KeyError: If the key does not exist.
    """
    context = safe_get_runner_context()
    context.external_context.delete(key)


def keys() -> KeysView[str]:
    """
    Get the keys of the context.

    Returns:
        KeysView[str]: The keys in the context.
    """
    context = safe_get_runner_context()
    return context.external_context.keys()


def set_config(
    *,
    timeout: float | None = None,
    end_on_error: bool | None = None,
    logging_setting: AllowableLogLevels | None = None,
    log_file: str | os.PathLike | None = None,
    broadcast_callback: (
        Callable[[str], None] | Callable[[str], Coroutine[None, None, None]] | None
    ) = None,
    prompt_injection: bool | None = None,
    save_state: bool | None = None,
):
    """
    Sets the global configuration for the executor. This will be propagated to all new runners created after this call.

    - If you call this function after the runner has been created, it will not affect the current runner.
    - This function will only overwrite the values that are provided, leaving the rest unchanged.


    """

    if is_context_active():
        warnings.warn(
            "The executor config is being set after the runner has been created, this is not recommended"
        )

    config = global_executor_config.get()

    if logging_setting or log_file:
        # default will be set at module import time, this is for overwrites
        configure_module_logging(level=logging_setting, log_file=log_file)

    new_config = config.precedence_overwritten(
        timeout=timeout,
        end_on_error=end_on_error,
        logging_setting=logging_setting,
        log_file=log_file,
        subscriber=broadcast_callback,
        prompt_injection=prompt_injection,
        save_state=save_state,
    )

    global_executor_config.set(new_config)


class RTContextLoggingAdapter(logging.LoggerAdapter):
    def process(self, msg, kwargs):
        try:
            parent_id = get_parent_id()
            run_id = get_run_id()
            session_id = get_session_id()

        except ContextError:
            parent_id = None
            run_id = None
            session_id = None

        new_variables = {
            "node_id": parent_id,
            "run_id": run_id,
            "session_id": session_id,
        }

        kwargs["extra"] = {**kwargs.get("extra", {}), **new_variables}

        return msg, kwargs


def session_id():
    """
    Gets the current session ID if it exists, otherwise returns None.
    """
    try:
        return get_session_id()
    except ContextError:
        return None



import base64
from unittest.mock import Mock, patch
from urllib.error import HTTPError

import pytest

from railtracks.llm.encoding import detect_source, encode, ensure_data_uri, _is_base64_image
from railtracks.llm.image_formats import detect_image_mime_from_bytes


class TestDetectSource:
    @pytest.mark.parametrize(
        "path,expected",
        [
            ("data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAUA", "data_uri"),
            ("data:image/jpeg;base64,/9j/4AAQSkZJRg", "data_uri"),
            ("http://example.com/image.png", "url"),
            ("https://example.com/image.png", "url"),
            ("ftp://example.com/image.png", "url"),
            ("ftps://example.com/image.png", "url"),
            ("/path/to/image.png", "local"),
            ("file:///path/to/image.png", "local"),
            ("images/photo.jpg", "local"),
            ("./local/image.png", "local"),
            ("../images/photo.jpg", "local"),
            ("~/images/photo.jpg", "local"),
        ],
    )
    def test_detect_source_types(self, path, expected):
        assert detect_source(path) == expected

    @patch("os.name", "nt")
    def test_detect_windows_forward_slash(self):
        assert detect_source("C:/Users/image.png") == "local"

    @patch("os.name", "nt")
    def test_detect_windows_backslash(self):
        assert detect_source("C:\\Users\\image.png") == "local"

    @patch("os.name", "nt")
    def test_detect_windows_unc_path(self):
        assert detect_source("\\\\server\\share\\image.png") == "local"

    def test_invalid_scheme_raises_error(self):
        with pytest.raises(ValueError, match="Could not determine image source type"):
            detect_source("invalid://example.com/image.png")

    def test_detect_source_relative_path_with_spaces(self):
        assert detect_source("images/my photo.jpg") == "local"


class TestEncode:
    def test_encode_local_file(self, tmp_path):
        image_data = b"fake image data"
        image_file = tmp_path / "test.png"
        image_file.write_bytes(image_data)

        result = encode(str(image_file))
        assert result == base64.b64encode(image_data).decode("utf-8")

    def test_file_not_found(self):
        with pytest.raises(FileNotFoundError, match="File not found"):
            encode("/nonexistent/file.png")

    def test_path_is_directory(self, tmp_path):
        directory = tmp_path / "test_dir"
        directory.mkdir()
        with pytest.raises(ValueError, match="Path is not a file"):
            encode(str(directory))

    @patch("railtracks.llm.encoding.request.urlopen")
    def test_encode_url(self, mock_urlopen):
        image_data = b"url image data"
        mock_response = Mock()
        mock_response.read.return_value = image_data
        mock_response.__enter__ = Mock(return_value=mock_response)
        mock_response.__exit__ = Mock(return_value=False)
        mock_urlopen.return_value = mock_response

        result = encode("https://example.com/image.png")
        assert result == base64.b64encode(image_data).decode("utf-8")

    @patch("railtracks.llm.encoding.request.urlopen")
    def test_url_http_error(self, mock_urlopen):
        mock_urlopen.side_effect = HTTPError("http://example.com", 404, "Not Found", None, None)  # type: ignore
        with pytest.raises(ValueError, match="Failed to encode URL"):
            encode("http://example.com/image.png")

    def test_data_uri_raises_error(self):
        with pytest.raises(ValueError, match="Data is already in byte64 encoded format"):
            encode("data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAUA")

    def test_empty_file_raises_error(self, tmp_path):
        empty_file = tmp_path / "empty.png"
        empty_file.write_bytes(b"")
        with pytest.raises(ValueError, match="Failed to encode image"):
            encode(str(empty_file))

    def test_encode_binary_roundtrip(self, tmp_path):
        binary_data = bytes(range(256))
        binary_file = tmp_path / "binary.dat"
        binary_file.write_bytes(binary_data)

        result = encode(str(binary_file))
        decoded = base64.b64decode(result)
        assert decoded == binary_data

    @pytest.mark.parametrize(
        "extension,header",
        [
            ("png", b"\x89PNG\r\n\x1a\n"),
            ("jpg", b"\xff\xd8\xff"),
            ("gif", b"GIF89a"),
            ("webp", b"RIFF"),
        ],
    )
    def test_different_formats(self, tmp_path, extension, header):
        image_file = tmp_path / f"image.{extension}"
        image_file.write_bytes(header)
        result = encode(str(image_file))
        assert result == base64.b64encode(header).decode("utf-8")

    def test_encode_non_image_binary(self, tmp_path):
        non_image_file = tmp_path / "random.bin"
        non_image_file.write_bytes(b"\x00\x01\x02\x03\x04\x05")
        result = encode(str(non_image_file))
        assert result == base64.b64encode(b"\x00\x01\x02\x03\x04\x05").decode("utf-8")


class TestAdditionalEncodingCases:
    def test_ensure_data_uri_valid_data_uri(self):
        uri = "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAUA"
        assert ensure_data_uri(uri).startswith("data:image/png;base64,")

    def test_ensure_data_uri_invalid_header(self):
        bad_uri = "data:image/png;base64iVBORw0KGgoAAAANSUhEUgAAAAUA"
        with pytest.raises(ValueError, match="Incomplete data URI"):
            ensure_data_uri(bad_uri)

    def test_ensure_data_uri_plain_base64(self):
        png_bytes = b"\x89PNG\r\n\x1a\n"
        b64 = base64.b64encode(png_bytes).decode("utf-8")
        result = ensure_data_uri(b64)
        assert result.startswith("data:image/png;base64,")

    def test_ensure_data_uri_invalid_base64(self):
        with pytest.raises(ValueError, match="Provided string is not valid base64"):
            ensure_data_uri("not_base64!!")

    def test_is_base64_image_invalid(self):
        assert not _is_base64_image("not_base64!!")

    def test_is_base64_image_valid_non_image(self):
        # base64 for "hello world" (not an image)
        b64 = base64.b64encode(b"hello world").decode("utf-8")
        assert not _is_base64_image(b64)

    def test_detect_image_mime_from_bytes_unsupported(self):
        # Should return None for random bytes
        assert detect_image_mime_from_bytes(b"abcdefg") is None




import pytest
from uuid import UUID
from railtracks.vector_stores.vector_store_base import (
    MetadataKeys,
    Metric,
    Chunk,
    SearchResult,
    FetchResult,
    SearchResponse,
    FetchResponse,
    OneOrMany,
    VectorStore,
)


class TestMetadataKeysEnum:
    """Tests for MetadataKeys enum."""

    def test_content_value(self):
        """Verify CONTENT enum value is correct."""
        assert MetadataKeys.CONTENT.value == "__content__"

    def test_document_value(self):
        """Verify DOCUMENT enum value is correct."""
        assert MetadataKeys.DOCUMENT.value == "__document__"

    def test_enum_members_by_name(self):
        """Verify enum members are accessible by name."""
        assert MetadataKeys["CONTENT"] == MetadataKeys.CONTENT
        assert MetadataKeys["DOCUMENT"] == MetadataKeys.DOCUMENT

    def test_enum_members_count(self):
        """Verify correct number of enum members."""
        assert len(list(MetadataKeys)) == 2


class TestMetricEnum:
    """Tests for Metric enum."""

    def test_cosine_value(self):
        """Verify cosine metric value."""
        assert Metric.cosine.value == "cosine"

    def test_l2_value(self):
        """Verify L2 metric value."""
        assert Metric.l2.value == "l2"

    def test_dot_value(self):
        """Verify dot metric value."""
        assert Metric.dot.value == "dot"

    def test_enum_members_count(self):
        """Verify correct number of enum members."""
        assert len(list(Metric)) == 3

    def test_enum_members_accessible_by_name(self):
        """Verify all metrics are accessible by name."""
        assert Metric["cosine"] == Metric.cosine
        assert Metric["l2"] == Metric.l2
        assert Metric["dot"] == Metric.dot

class TestSearchResultDataclass:
    """Tests for SearchResult dataclass."""

    def test_create_search_result_with_all_fields(self, simple_content, simple_document, short_vector, small_distance, simple_metadata):
        """Create SearchResult with all fields."""
        result = SearchResult(
            id="id1",
            distance=small_distance,
            content=simple_content,
            vector=short_vector,
            document=simple_document,
            metadata=simple_metadata
        )
        assert result.id == "id1"
        assert result.distance == small_distance
        assert result.content == simple_content
        assert result.vector == short_vector
        assert result.document == simple_document
        assert result.metadata == simple_metadata

    def test_create_search_result_with_required_fields(self, simple_content, short_vector, small_distance):
        """Create SearchResult with only required fields."""
        result = SearchResult(
            id="id1",
            distance=small_distance,
            content=simple_content,
            vector=short_vector
        )
        assert result.id == "id1"
        assert result.distance == small_distance
        assert result.content == simple_content
        assert result.vector == short_vector
        assert result.document is None
        assert result.metadata == {}

    def test_search_result_with_zero_distance(self, simple_content, short_vector, zero_distance):
        """Create SearchResult with zero distance."""
        result = SearchResult(
            id="id1",
            distance=zero_distance,
            content=simple_content,
            vector=short_vector
        )
        assert result.distance == zero_distance

    def test_search_result_with_large_vector(self, simple_content, large_distance, long_vector):
        """Create SearchResult with large vector."""
        result = SearchResult(
            id="id1",
            distance=large_distance,
            content=simple_content,
            vector=long_vector
        )
        assert len(result.vector) == 1000


class TestFetchResultDataclass:
    """Tests for FetchResult dataclass."""

    def test_create_fetch_result_with_all_fields(self, simple_content, simple_document, short_vector, simple_metadata):
        """Create FetchResult with all fields."""
        result = FetchResult(
            id="id1",
            content=simple_content,
            vector=short_vector,
            document=simple_document,
            metadata=simple_metadata
        )
        assert result.id == "id1"
        assert result.content == simple_content
        assert result.vector == short_vector
        assert result.document == simple_document
        assert result.metadata == simple_metadata

    def test_create_fetch_result_with_required_fields(self, simple_content, short_vector):
        """Create FetchResult with only required fields."""
        result = FetchResult(
            id="id1",
            content=simple_content,
            vector=short_vector
        )
        assert result.id == "id1"
        assert result.content == simple_content
        assert result.vector == short_vector
        assert result.document is None
        assert result.metadata == {}

    def test_fetch_result_with_custom_metadata(self, simple_content, simple_vector, complex_metadata):
        """Create FetchResult with custom metadata."""
        result = FetchResult(
            id="id1",
            content=simple_content,
            vector=simple_vector,
            metadata=complex_metadata
        )
        assert result.metadata == complex_metadata


class TestTypeAliases:
    """Tests for type aliases."""

    def test_search_response_is_list_of_search_result(self, simple_content, simple_vector, small_distance, large_distance):
        """Verify SearchResponse is a list of SearchResult."""
        result1 = SearchResult(id="1", distance=small_distance, content=simple_content, vector=simple_vector)
        result2 = SearchResult(id="2", distance=large_distance, content=simple_content, vector=simple_vector)
        
        response: SearchResponse = [result1, result2]
        
        assert isinstance(response, list)
        assert len(response) == 2
        assert all(isinstance(r, SearchResult) for r in response)

    def test_fetch_response_is_list_of_fetch_result(self, simple_content, simple_vector):
        """Verify FetchResponse is a list of FetchResult."""
        result1 = FetchResult(id="1", content=simple_content, vector=simple_vector)
        result2 = FetchResult(id="2", content=simple_content, vector=simple_vector)
        
        response: FetchResponse = [result1, result2]
        
        assert isinstance(response, list)
        assert len(response) == 2
        assert all(isinstance(r, FetchResult) for r in response)

    def test_one_or_many_with_single_value(self):
        """Test OneOrMany with single value."""
        value: OneOrMany[str] = "single"
        assert isinstance(value, str)

    def test_one_or_many_with_list(self):
        """Test OneOrMany with list value."""
        value: OneOrMany[str] = ["first", "second"]
        assert isinstance(value, list)
        assert len(value) == 2


class TestVectorStoreAbstractClass:
    """Tests for VectorStore abstract base class."""

    def test_cannot_instantiate_abstract_class(self):
        """Verify cannot instantiate abstract VectorStore directly."""
        with pytest.raises(TypeError, match="Can't instantiate abstract class"):
            store = VectorStore(  # type: ignore[abstract]
                collection_name="test",
                embedding_function=lambda x: [[0.1]]
            )

    def test_all_abstract_methods_defined(self):
        """Verify all abstract methods are defined."""
        abstract_methods = VectorStore.__abstractmethods__
        
        expected_methods = {"upsert", "fetch", "search", "delete", "count"}
        assert abstract_methods == expected_methods

    def test_concrete_subclass_can_be_created(self):
        """Verify concrete subclass can be created by implementing all methods."""
        
        class ConcreteVectorStore(VectorStore):
            def upsert(self, content):
                return []
            
            def fetch(self, ids):
                return []
            
            def search(self, query, top_k=10, where=None, include=None):
                return []
            
            def delete(self, ids, where=None):
                pass
            
            def count(self):
                return 0
        
        # Should not raise
        store = ConcreteVectorStore(
            collection_name="test",
            embedding_function=lambda x: [[0.1]]
        )
        assert store._collection_name == "test"



from typing import Literal, Type, TypeVar

from railtracks.built_nodes.concrete.response import LLMResponse

from ..built_nodes.concrete._llm_base import LLMBase
from ..human_in_the_loop import ChatUI, HILMessage
from ..human_in_the_loop.local_chat_ui import UserMessageAttachment
from ..llm.history import MessageHistory
from ..llm.message import AssistantMessage, UserMessage
from ..utils.logging.create import get_rt_logger
from ._call import call

logger = get_rt_logger("Interactive")

_TOutput = TypeVar("_TOutput", bound=LLMResponse)


def _process_attachment(attachments: list[UserMessageAttachment]) -> list[str]:
    """Processes a list of attachments and returns their data or URLs.

    Args:
        attachments: A list of UserMessageAttachment objects.

    Returns:
        A list of strings containing the processed data or URLs.
    """
    processed = []
    for attachment in attachments:
        if attachment.type == "file":
            processed.append(attachment.data)
        elif attachment.type == "url":
            processed.append(attachment.url)
    return processed


async def _chat_ui_interactive(
    chat_ui: ChatUI,
    node: Type[LLMBase[_TOutput, _TOutput, Literal[False]]],
    initial_message_to_user: str | None,
    initial_message_to_agent: str | None,
    turns: int | None,
    *args,
    **kwargs,
) -> _TOutput:
    """Handles the interactive session logic using the ChatUI interface.
    Args:
        chat_ui: An instance of the ChatUI class to manage the user interface.
        node: The LLMBase class to interact with.
        initial_message_to_user: An optional message to display to the user at the start of the chat session.
        initial_message_to_agent: An optional message to send to the agent to initiate the conversation.
        turns: The maximum number of conversational turns before the session terminates. If None,
               the session continues until manually closed.
        *args: Additional positional arguments to pass to the node constructor.
        **kwargs: Additional keyword arguments to pass to the node constructor.
    Returns:
        The final output from the node after the interactive session concludes.
    """
    msg_history = MessageHistory([])

    if initial_message_to_user is not None:
        await chat_ui.send_message(HILMessage(content=initial_message_to_user))
        msg_history.append(AssistantMessage(content=initial_message_to_user))
    if initial_message_to_agent is not None:
        msg_history.append(UserMessage(content=initial_message_to_agent))

    last_tool_idx = 0  # To track the last processed tool response, not sure how efficient this makes things

    while chat_ui.is_connected:
        message = await chat_ui.receive_message()
        if message is None:
            continue  # could be `break` but I want to ensure chat_ui.is_connected is updated properly

        attachments = []
        if message.attachments is not None:
            attachments = _process_attachment(message.attachments)
        msg_history.append(UserMessage(content=message.content, attachment=attachments))

        response = await call(node, msg_history, *args, **kwargs)

        msg_history = response.message_history.copy()

        await chat_ui.send_message(HILMessage(content=response.content))
        await chat_ui.update_tools(response.tool_invocations[last_tool_idx:])

        last_tool_idx = len(response.tool_invocations)
        if turns is not None:
            turns -= 1
            if turns <= 0:
                await chat_ui.disconnect()

    logger.info("Ended Local Chat Session")

    return response  # type: ignore


async def local_chat(
    node: type[LLMBase[_TOutput, _TOutput, Literal[False]]],
    initial_message_to_user: str | None = None,
    initial_message_to_agent: str | None = None,
    turns: int | None = None,
    port: int | None = None,
    host: str | None = None,
    auto_open: bool | None = True,
    *args,
    **kwargs,
) -> _TOutput:
    """Starts an interactive session with an LLM-based agent.

    This function launches a local web server, providing a chat interface for
    real-time interaction with a specified `LLMBase` node. It facilitates a
    turn-by-turn conversation with the agent.

    Args:
        node: The `LLMBase` class to interact with.
        initial_message_to_user: An optional message to display to the user
            at the start of the chat session.
        initial_message_to_agent: An optional message to send to the agent to
            initiate the conversation.
        turns: The maximum number of conversational turns before the session
            terminates. If `None`, the session continues until manually closed.
        port: The network port for the web server. If `None`, a random
            available port is selected.
        host: The network host for the web server. Defaults to 'localhost'.
        auto_open: If `True`, automatically opens the chat interface in a
            web browser.
        *args: Additional positional arguments to pass to the `node` constructor.
        **kwargs: Additional keyword arguments to pass to the `node` constructor.

    Returns:
        The final output from the node after the interactive session concludes.
        The return type matches the node's `_TOutput` generic type.
    """

    chat_ui_kwargs = {}
    if port is not None:
        chat_ui_kwargs["port"] = port
    if host is not None:
        chat_ui_kwargs["host"] = host
    if auto_open is not None:
        chat_ui_kwargs["auto_open"] = auto_open

    if not issubclass(node, LLMBase):
        raise ValueError(
            "Interactive sessions only support nodes that are children of LLMBase."
        )
    response = None
    try:
        logger.info("Connecting with Local Chat Session")

        chat_ui = ChatUI(**chat_ui_kwargs)

        await chat_ui.connect()

        response = await _chat_ui_interactive(
            chat_ui,
            node,
            initial_message_to_user,
            initial_message_to_agent,
            turns,
            *args,
            **kwargs,
        )

    except Exception as e:
        logger.error(f"Error during interactive session: {e}")
    finally:
        return response  # type: ignore



from .chroma import ChromaVectorStore
from .chunking.base_chunker import Chunk
from .chunking.fixed_token_chunker import FixedTokenChunker
from .chunking.media_parser import MediaParser

__all__ = ["ChromaVectorStore", "Chunk", "FixedTokenChunker", "MediaParser"]



from .._exception_base import RTLLMError
from ..history import MessageHistory


class ModelError(RTLLMError):
    """
    Any Large Language Model (LLM) error.
    """

    def __init__(
        self,
        reason: str,
        message_history: MessageHistory = None,
    ):
        self.reason = reason
        self.message_history = message_history

        message = f"{self._color('Failure reason: ', self.BOLD_RED)}{self._color(reason, self.RED)}"
        super().__init__(message)

    def __str__(self):
        base = super().__str__()
        details = []
        if self.message_history:
            mh_str = str(self.message_history)
            indented_mh = "\n".join(
                "    " + line for line in mh_str.splitlines()
            )  # 2 indents (2-spaces) per indent
            details.append(
                self._color("Message History:\n", self.BOLD_GREEN)
                + self._color(indented_mh, self.GREEN)
            )
        if details:
            notes_str = (
                "\n"
                + self._color("Details:\n", self.BOLD_GREEN)
                + "\n".join(f"  {d}" for d in details)
            )
            return f"\n{self._color(base, self.RED)}{notes_str}"
        return self._color(base, self.RED)


class ModelNotFoundError(RTLLMError):
    def __init__(self, reason: str, notes: list[str] = None):
        self.reason = reason
        self.notes = notes or []
        super().__init__(reason)

    def __str__(self):
        base = super().__str__()
        if self.notes:
            notes_str = (
                "\n"
                + self._color("Tips to debug:\n", self.GREEN)
                + "\n".join(self._color(f"- {note}", self.GREEN) for note in self.notes)
            )
            return f"\n{self._color(base, self.RED)}{notes_str}"
        return self._color(base, self.RED)


class FunctionCallingNotSupportedError(ModelError):
    """Error raised when a model does not support function calling."""

    def __init__(self, model_name: str):
        super().__init__(
            reason=f"Model {model_name} does not support function calling. Chat with tools is not supported."
        )



import pytest
import railtracks as rt
from typing import List, Callable, Type
from pydantic import BaseModel, Field
from railtracks.llm import SystemMessage
import random 

# ============ System Messages ===========
@pytest.fixture
def encoder_system_message():
    return SystemMessage("You are a text encoder. Encode the input string into bytes and do a random operation on them. You can use the following operations: reverse the byte order, or repeat each byte twice, or jumble the bytes.")


@pytest.fixture
def decoder_system_message():
    return SystemMessage("You are a text decoder. Decode the bytes into a string.")


# ============ Helper function for test_function.py ===========
@pytest.fixture
def _agent_node_factory():
    """
    Returns a top level agent node with mock model for testing 
    """

    def _create_node(test_function: Callable, llm: rt.llm.ModelBase):
        """
        Creates a top-level node for testing function nodes.

        Args:
            test_function: The function to test.
            model_provider: The model provider to use (default: "openai").

        Returns:
            A ToolCallLLM node that can be used to test the function.
        """

        return rt.agent_node(
            name=f"TestNode-{test_function.__name__}",
            system_message=SystemMessage(
                f"You are a test node for the function {test_function.__name__}"
            ),
            llm=llm,
            tool_nodes={rt.function_node(test_function)},
        )

    return _create_node


# ============ Output Models ===========
class SimpleOutput(BaseModel):  # simple structured output case
    text: str = Field(description="The text to return")
    number: int = Field(description="The number to return")


@pytest.fixture
def simple_output_model():
    return SimpleOutput

# =====================================================

# ============ Context Variables ===========
@pytest.fixture
def _reset_tools_called():
    def _reset(val=0):
        rt.context.put("tools_called", val)
    return _reset

@pytest.fixture
def _increment_tools_called():
    """Increments the tools_called context variable by 1"""
    def _increment():
        count = rt.context.get("tools_called", 0)
        rt.context.put("tools_called", count + 1)
    return _increment